from __future__ import annotations

import datetime
from dataclasses import dataclass
from typing import TYPE_CHECKING

import pytest
from datadoc_model.all_optional.model import DataSetState

from dapla_metadata.datasets.dapla_dataset_path_info import ISO_YEAR
from dapla_metadata.datasets.dapla_dataset_path_info import ISO_YEAR_MONTH
from dapla_metadata.datasets.dapla_dataset_path_info import ISO_YEAR_MONTH_DAY
from dapla_metadata.datasets.dapla_dataset_path_info import SSB_BIMESTER
from dapla_metadata.datasets.dapla_dataset_path_info import DaplaDatasetPathInfo
from tests.datasets.constants import TEST_BUCKET_PARQUET_FILEPATH_WITH_SHORTNAME
from tests.datasets.constants import TEST_PARQUET_FILEPATH

if TYPE_CHECKING:
    import pathlib


@dataclass
class DatasetPathTestCase:
    """Structure to define attributes needed for a test case."""

    path: str
    expected_contains_data_from: datetime.date
    expected_contains_data_until: datetime.date


TEST_CASES = [
    DatasetPathTestCase(
        path="grensehandel_imputert_p2022-10-01_p2022-12-31_v1.parquet",
        expected_contains_data_from=datetime.date(2022, 10, 1),
        expected_contains_data_until=datetime.date(2022, 12, 31),
    ),
    DatasetPathTestCase(
        path="grensehandel_imputert_p2022-10_p2022-12_v1.parquet",
        expected_contains_data_from=datetime.date(2022, 10, 1),
        expected_contains_data_until=datetime.date(2022, 12, 31),
    ),
    DatasetPathTestCase(
        path="flygende_objekter_p2019_v1.parquet",
        expected_contains_data_from=datetime.date(2019, 1, 1),
        expected_contains_data_until=datetime.date(2019, 12, 31),
    ),
    DatasetPathTestCase(
        path="framskrevne-befolkningsendringer_p2019_p2050_v1.parquet",
        expected_contains_data_from=datetime.date(2019, 1, 1),
        expected_contains_data_until=datetime.date(2050, 12, 31),
    ),
    DatasetPathTestCase(
        path="ufo_observasjoner_p2019_p2020_v1.parquet",
        expected_contains_data_from=datetime.date(2019, 1, 1),
        expected_contains_data_until=datetime.date(2020, 12, 31),
    ),
    DatasetPathTestCase(
        path="omsetning_p2020W15_v1.parquet",
        expected_contains_data_from=datetime.date(2020, 4, 6),
        expected_contains_data_until=datetime.date(2020, 4, 12),
    ),
    DatasetPathTestCase(
        path="omsetning_p1981-W52_v1.parquet",
        expected_contains_data_from=datetime.date(1981, 12, 21),
        expected_contains_data_until=datetime.date(1981, 12, 27),
    ),
    DatasetPathTestCase(
        path="personinntekt_p2022H1_v1.parquet",
        expected_contains_data_from=datetime.date(2022, 1, 1),
        expected_contains_data_until=datetime.date(2022, 6, 30),
    ),
    DatasetPathTestCase(
        path="personinntekt_p2022-H1_v1.parquet",
        expected_contains_data_from=datetime.date(2022, 1, 1),
        expected_contains_data_until=datetime.date(2022, 6, 30),
    ),
    DatasetPathTestCase(
        path="nybilreg_p2022T1_v1.parquet",
        expected_contains_data_from=datetime.date(2022, 1, 1),
        expected_contains_data_until=datetime.date(2022, 4, 30),
    ),
    DatasetPathTestCase(
        path="nybilreg_p2022-T1_v1.parquet",
        expected_contains_data_from=datetime.date(2022, 1, 1),
        expected_contains_data_until=datetime.date(2022, 4, 30),
    ),
    DatasetPathTestCase(
        path="varehandel_p2018Q1_p2018Q4_v1.parquet",
        expected_contains_data_from=datetime.date(2018, 1, 1),
        expected_contains_data_until=datetime.date(2018, 12, 31),
    ),
    DatasetPathTestCase(
        path="varehandel_p2018-Q1_p2018-Q4_v1.parquet",
        expected_contains_data_from=datetime.date(2018, 1, 1),
        expected_contains_data_until=datetime.date(2018, 12, 31),
    ),
    DatasetPathTestCase(
        path="pensjon_p2018Q1_v1.parquet",
        expected_contains_data_from=datetime.date(2018, 1, 1),
        expected_contains_data_until=datetime.date(2018, 3, 31),
    ),
    DatasetPathTestCase(
        path="pensjon_p2018-Q1_v1.parquet",
        expected_contains_data_from=datetime.date(2018, 1, 1),
        expected_contains_data_until=datetime.date(2018, 3, 31),
    ),
    DatasetPathTestCase(
        path="skipsanloep_p2021B2_v1.parquet",
        expected_contains_data_from=datetime.date(2021, 3, 1),
        expected_contains_data_until=datetime.date(2021, 4, 30),
    ),
    DatasetPathTestCase(
        path="skipsanloep_p2021-B2_v1.parquet",
        expected_contains_data_from=datetime.date(2021, 3, 1),
        expected_contains_data_until=datetime.date(2021, 4, 30),
    ),
    DatasetPathTestCase(
        path="skipsanloep_p2022B1_v1.parquet",
        expected_contains_data_from=datetime.date(2022, 1, 1),
        expected_contains_data_until=datetime.date(2022, 2, 28),
    ),
    DatasetPathTestCase(
        path="skipsanloep_p2022-B1_v1.parquet",
        expected_contains_data_from=datetime.date(2022, 1, 1),
        expected_contains_data_until=datetime.date(2022, 2, 28),
    ),
]


@pytest.fixture(
    ids=[tc.path for tc in TEST_CASES],
    params=TEST_CASES,
)
def test_data(request: pytest.FixtureRequest) -> DatasetPathTestCase:
    return request.param


@pytest.fixture
def dataset_path(test_data: DatasetPathTestCase) -> DaplaDatasetPathInfo:
    return DaplaDatasetPathInfo(test_data.path)


@pytest.fixture
def expected_contains_data_from(test_data: DatasetPathTestCase) -> datetime.date:
    return test_data.expected_contains_data_from


@pytest.fixture
def expected_contains_data_until(test_data: DatasetPathTestCase) -> datetime.date:
    return test_data.expected_contains_data_until


def test_extract_period_info_date_from(
    dataset_path: DaplaDatasetPathInfo,
    expected_contains_data_from: datetime.date,
):
    assert dataset_path.contains_data_from == expected_contains_data_from


def test_extract_period_info_date_until(
    dataset_path: DaplaDatasetPathInfo,
    expected_contains_data_until: datetime.date,
):
    assert dataset_path.contains_data_until == expected_contains_data_until


@pytest.mark.parametrize(
    "data",
    [
        "nonsen.data",
        "nonsens2.parquet",
        TEST_PARQUET_FILEPATH.name,
    ],
)
def test_extract_period_info_no_period_info_in_path(data: str):
    assert DaplaDatasetPathInfo(data).contains_data_from is None


@pytest.mark.parametrize(
    ("path_parts_to_insert", "expected_result"),
    [
        ("kildedata", DataSetState.SOURCE_DATA),
        ("inndata", DataSetState.INPUT_DATA),
        ("roskildedata/klargjorte-data", DataSetState.PROCESSED_DATA),
        ("klargjorte_data", DataSetState.PROCESSED_DATA),
        ("klargjorte-data", DataSetState.PROCESSED_DATA),
        ("statistikk", DataSetState.STATISTICS),
        ("", None),
    ],
)
def test_get_dataset_state(
    full_dataset_state_path: pathlib.Path,
    expected_result: DataSetState,
):
    actual_state = DaplaDatasetPathInfo(full_dataset_state_path).dataset_state
    assert actual_state == expected_result


@pytest.mark.parametrize(
    ("path", "expected"),
    [
        ("person_data_v1", "1"),
        ("person_data_v2", "2"),
        ("person_data_vwrong", None),
        ("person_data", None),
        ("person_testdata_p2021-12-31_p2021-12-31_v20", "20"),
    ],
)
def test_get_dataset_version(
    path: str,
    expected: str | None,
):
    assert DaplaDatasetPathInfo(path).dataset_version == expected


# These tests covers both date until after date from, mix of SSB keys and invalid SSB keys
@pytest.mark.parametrize(
    "dataset_path_name",
    [
        "ufo_observasjoner_p2019_p1920_v1.parquet",
        "varehandel_p2018H2_p2018H1_v1.parquet",
        "varehandel_p2018Q1_p2018H2_v1.parquet",
        "sykkeltransport_p1973B8_v1.parquet",
    ],
)
def test_extract_period_info_date_from_invalid_pathname(dataset_path_name: str) -> None:
    dataset = DaplaDatasetPathInfo(dataset_path_name)
    assert dataset.contains_data_from is None


@pytest.mark.parametrize(
    "dataset_path_name",
    [
        "ufo_observasjoner_p2019_p1920_v1.parquet",
        "varehandel_p2018H2_p2018H1_v1.parquet",
        "varehandel_p2018Q1_p2018H2_v1.parquet",
        "sykkeltransport_p1973B2_p2020T8_v1.parquet",
    ],
)
def test_extract_period_info_date_until_invalid_pathname(
    dataset_path_name: str,
) -> None:
    dataset = DaplaDatasetPathInfo(dataset_path_name)
    assert dataset.contains_data_until is None


@pytest.mark.parametrize(
    ("date_format", "period"),
    [
        (ISO_YEAR, "1980"),
        (ISO_YEAR_MONTH, "1888-11"),
        (ISO_YEAR_MONTH_DAY, "2203-01-24"),
        (SSB_BIMESTER, "1963B3"),
    ],
)
def test_date_format_return_date_object_period_start(date_format, period):
    assert isinstance(date_format.get_floor(period), datetime.date)


@pytest.mark.parametrize(
    ("date_format", "period"),
    [
        (ISO_YEAR, "1980"),
        (ISO_YEAR_MONTH, "1888-11"),
        (ISO_YEAR_MONTH_DAY, "2203-01-24"),
        (SSB_BIMESTER, "1963B3"),
    ],
)
def test_date_format_return_date_object_period_end(date_format, period):
    assert isinstance(date_format.get_ceil(period), datetime.date)


@pytest.mark.parametrize(
    ("date_format", "period", "expected"),
    [
        (ISO_YEAR, "1980", datetime.date(1980, 1, 1)),
        (ISO_YEAR_MONTH, "1888-11", datetime.date(1888, 11, 1)),
        (ISO_YEAR_MONTH_DAY, "2203-01-24", datetime.date(2203, 1, 24)),
        (SSB_BIMESTER, "1963B3", datetime.date(1963, 5, 1)),
    ],
)
def test_date_format_correct_from_date(date_format, period, expected: datetime.date):
    assert date_format.get_floor(period) == expected


@pytest.mark.parametrize(
    ("date_format", "period", "expected"),
    [
        (ISO_YEAR, "1980", datetime.date(1980, 12, 31)),
        (ISO_YEAR_MONTH, "1888-11", datetime.date(1888, 11, 30)),
        (ISO_YEAR_MONTH_DAY, "2203-01-24", datetime.date(2203, 1, 24)),
        (SSB_BIMESTER, "1963B3", datetime.date(1963, 6, 30)),
    ],
)
def test_date_format_correct_end_date(date_format, period, expected):
    assert date_format.get_ceil(period) == expected


@pytest.mark.parametrize(
    ("data", "expected"),
    [
        (TEST_BUCKET_PARQUET_FILEPATH_WITH_SHORTNAME, "befolkning"),
        (
            "gs://ssb-staging-dapla-felles-data-delt/datadoc/person_data_v1.parquet",
            "datadoc",
        ),
        ("inndata/person_data_v1.parquet", None),
        ("stat/inndata/person_data", "stat"),
        ("buckets/stat/inndata/person_data", None),
        ("buckets/bucket_name/stat/inndata/person_data", "stat"),
        (
            "gs://statistikk/produkt/klargjorte-data/persondata_p1990-Q1_p2023-Q4_v1/aar=2019/data.parquet",
            "produkt",
        ),
        (
            "gs://statistikk/produkt/persondata_p1990-Q1_p2023-Q4_v1/aar=2019/data.parquet",
            None,
        ),
        (
            "buckets/ssb-staging-dapla-felles-data-delt/person_data_p2021_v2.parquet",
            None,
        ),
    ],
)
def test_extract_shortname_in_path(data: str, expected: str):
    assert DaplaDatasetPathInfo(data).statistic_short_name == expected


@pytest.mark.parametrize(
    ("data", "expected"),
    [
        (
            TEST_BUCKET_PARQUET_FILEPATH_WITH_SHORTNAME,
            "ssb-staging-dapla-felles-data-delt",
        ),
        (
            "gs://datadoc/person_data_v1.parquet",
            "datadoc",
        ),
        (
            "buckets/ssb-bucket/stat_name/klargjorte_data/person_data_p2021-12-31_p2021-12-31_v1.parquet",
            "ssb-bucket",
        ),
    ],
)
def test_extract_bucketname_in_path(data: str, expected: str):
    assert DaplaDatasetPathInfo(data).bucket_name == expected


@pytest.mark.parametrize(
    ("data"),
    [
        "gs://ssb-staging-dapla-felles-data-delt/person_data_p2022_v1.parquet",
        "gs://ssb-staging-dapla-felles-data-delt/datadoc/person_data_v1.parquet",
        "gs://ssb-staging-dapla-felles-data-delt/datadoc/person_data_p2021_v3.parquet",
        "gs://ssb-staging-dapla-felles-data-delt/datadoc/utdata/person_data_v1.parquet",
        "gs://ssb-staging-dapla-felles-data-delt/datadoc/utdata/person_data_p2021.parquet",
        "gs://utdata/person_data_p2021_p2022_v2.parquet",
        "buckets/klargjorte_data/person_data_p2021-12-31_p2021-12-31_v1.parquet",
    ],
)
def test_path_complies_with_naming_standard_invalid_input(data: str):
    assert DaplaDatasetPathInfo(data).path_complies_with_naming_standard() is False


@pytest.mark.parametrize(
    ("data"),
    [
        "gs://ssb-staging-dapla-felles-data-delt/datadoc/utdata/person_data_p2021_v2.parquet",
        "gs://ssb-staging-dapla-felles-data-delt/datadoc/utdata/person_data_p2021_p2022_v2.parquet",
        "gs://ssb-staging-dapla-felles-data-delt/datadoc/utdata/undermappe/person_data_p2021_v2.parquet",
        "buckets/bucket_name/dataset/klargjorte_data/person_data_p2021-12-31_p2021-12-31_v1.parquet",
        "dataset/klargjorte_data/person_data_p2021-12-31_p2021-12-31_v1.parquet",
    ],
)
def test_path_complies_with_naming_standard_valid_input(data: str):
    assert DaplaDatasetPathInfo(data).path_complies_with_naming_standard() is True


def test_extract_shortname_in_path_invalid_gs_url():
    with pytest.raises(ValueError, match="non key-like path provided"):
        _ = DaplaDatasetPathInfo(
            "gs:/ssb-staging-dapla-felles-data-delt/datadoc/person_data_v1.parquet"
        ).dataset_short_name



import asyncio
import logging
import os
import time
from collections.abc import AsyncGenerator

from dapla_metadata.standards.name_validator import NamingStandardReport
from dapla_metadata.standards.name_validator import ValidationResult
from dapla_metadata.standards.name_validator import validate_directory

logger = logging.getLogger(__name__)


async def check_naming_standard(
    file_path: str | os.PathLike[str],
) -> list[ValidationResult]:
    """Check whether a given path follows the SSB naming standard.

    This function checks whether the provided `file_path` and subdirectories thereof comply
    with the naming standard. Currently we only examine '.parquet' files. Other files are ignored.

    Args:
        file_path: The path to a bucket, directory, or specific file to validate.
                This can be in the following forms:
                - A bucket URL in the form 'gs://ssb-dapla-felles-data-produkt-test'
                - An absolute path to a mounted bucket in the form '/buckets/produkt'
                - Any subdirectory or file thereof

                We also accept paths which don't yet exist so that you can test if a path will comply.

    Returns:
        list[ValidationResult]: A list of validation results,
        including success status, checked file path, messages, and any detected violations.

    Examples:
        >>> check_naming_standard("/data/example_file.parquet").success
        False

        >>> check_naming_standard("/buckets/produkt/datadoc/utdata/person_data_p2021_v2.parquet").success
        True
    """
    results = []

    # Begin validation.
    # For each file this returns a task which we can wait on to complete.
    # For each directory this returns another AsyncGenerator which must be unpacked below
    tasks: list[asyncio.Task] = []
    async for t in flatten_generator(validate_directory(str(file_path))):
        tasks.append(t)  # noqa: PERF401

    # 5 minute timeout for safety
    start_time = time.time()
    while time.time() < start_time + (5 * 60):
        for item in tasks:
            if isinstance(item, AsyncGenerator):
                # Drill down into lower directories to get the validation tasks from them
                tasks.remove(item)
                new_tasks = [t async for t in item]
                logger.debug("New Tasks: %s %s", len(new_tasks), new_tasks)
                tasks.extend(
                    new_tasks,
                )
            elif isinstance(item, asyncio.Task):
                if item.done():
                    logger.info("Validated %s", item.get_name())
                    tasks.remove(item)
                    results.append(item.result())

        logger.debug("Tasks: %s %s", len(tasks), tasks)
        logger.debug("Results: %s", len(results))

        if len(tasks) == 0:
            logger.info("Completed validation")
            break

        # Allow time for other processing to be performed
        await asyncio.sleep(0.001)

    return results


async def flatten_generator(gen: AsyncGenerator) -> AsyncGenerator[asyncio.Task, None]:
    """Recursively flatten nested async generators."""
    async for item in gen:
        if isinstance(item, AsyncGenerator):
            async for sub_item in flatten_generator(item):
                yield sub_item
        else:
            yield item


def generate_validation_report(
    validation_results: list[ValidationResult],
) -> NamingStandardReport:
    """Generate and print a formatted naming standard validation report.

    This function takes a list of `ValidationResult` objects, creates a
    `NamingStandardReport` instance, and prints the generated report.

    Args:
        validation_results: A list of ValidationResult objects that
        contain the outcomes of the name standard checks.

    Returns:
        NamingStandardReport: An instance of `NamingStandardReport` containing
        the validation results.
    """
    report = NamingStandardReport(validation_results=validation_results)
    print(report.generate_report())  # noqa: T201
    return report



"""Tests for the DatasetParser class."""

import io

import pandas as pd
import pytest
from datadoc_model.all_optional.model import DataType
from datadoc_model.all_optional.model import LanguageStringType
from datadoc_model.all_optional.model import LanguageStringTypeItem
from datadoc_model.all_optional.model import Variable
from upath import UPath

from dapla_metadata.datasets.dataset_parser import KNOWN_BOOLEAN_TYPES
from dapla_metadata.datasets.dataset_parser import KNOWN_DATETIME_TYPES
from dapla_metadata.datasets.dataset_parser import KNOWN_FLOAT_TYPES
from dapla_metadata.datasets.dataset_parser import KNOWN_INTEGER_TYPES
from dapla_metadata.datasets.dataset_parser import KNOWN_STRING_TYPES
from dapla_metadata.datasets.dataset_parser import DatasetParser
from dapla_metadata.datasets.dataset_parser import DatasetParserParquet
from tests.datasets.constants import TEST_PARQUET_FILEPATH
from tests.datasets.constants import TEST_PARQUET_GZIP_FILEPATH
from tests.datasets.constants import TEST_SAS7BDAT_FILEPATH


def test_use_abstract_class_directly():
    with pytest.raises(TypeError):
        DatasetParser().get_fields()


@pytest.mark.parametrize(
    "local_parser",
    [
        DatasetParser.for_file(TEST_PARQUET_FILEPATH),
        DatasetParser.for_file(TEST_PARQUET_GZIP_FILEPATH),
    ],
)
def test_get_fields_parquet(local_parser: DatasetParserParquet):
    expected_fields = [
        Variable(short_name="pers_id", data_type=DataType.STRING),
        Variable(short_name="tidspunkt", data_type=DataType.DATETIME),
        Variable(short_name="sivilstand", data_type=DataType.STRING),
        Variable(short_name="alm_inntekt", data_type=DataType.INTEGER),
        Variable(short_name="sykepenger", data_type=DataType.INTEGER),
        Variable(short_name="ber_bruttoformue", data_type=DataType.INTEGER),
        Variable(short_name="fullf_utdanning", data_type=DataType.STRING),
        Variable(short_name="hoveddiagnose", data_type=DataType.STRING),
    ]

    assert local_parser.get_fields() == expected_fields


def test_get_fields_sas7bdat():
    expected_fields = [
        Variable(
            short_name="tekst",
            name=LanguageStringType(
                [LanguageStringTypeItem(languageCode="nb", languageText="Tekst")],
            ),
            data_type=DataType.STRING,
        ),
        Variable(
            short_name="tall",
            name=LanguageStringType(
                [LanguageStringTypeItem(languageCode="nb", languageText="Tall")],
            ),
            data_type=DataType.FLOAT,
        ),
        Variable(
            short_name="dato",
            name=LanguageStringType(
                [LanguageStringTypeItem(languageCode="nb", languageText="Dato")],
            ),
            data_type=DataType.DATETIME,
        ),
    ]

    reader = DatasetParser.for_file(TEST_SAS7BDAT_FILEPATH)
    fields = reader.get_fields()

    assert fields == expected_fields


@pytest.mark.parametrize("file", ["my_dataset.csv", "my_dataset.xlsx", "my_dataset"])
def test_dataset_parser_unsupported_files(file: UPath):
    with pytest.raises(NotImplementedError):
        DatasetParser.for_file(UPath(file))


def test_transform_datatype_unknown_type():
    assert DatasetParser.transform_data_type("definitely not a known data type") is None


@pytest.mark.parametrize(
    ("expected", "concrete_type"),
    [
        *[(DataType.INTEGER, i) for i in KNOWN_INTEGER_TYPES],
        *[(DataType.FLOAT, i) for i in KNOWN_FLOAT_TYPES],
        *[(DataType.STRING, i) for i in KNOWN_STRING_TYPES],
        *[(DataType.DATETIME, i) for i in KNOWN_DATETIME_TYPES],
        *[(DataType.BOOLEAN, i) for i in KNOWN_BOOLEAN_TYPES],
    ],
)
def test_transform_datatype(expected: DataType, concrete_type: str):
    actual = DatasetParser.transform_data_type(concrete_type)
    assert actual == expected


@pytest.fixture
def parquet_with_index_column(tmp_path):
    """Create a parquet file with a column called __index_level_0__."""
    test_data = pd.read_csv(
        io.StringIO(
            """a	b
1	4
2	5
3	6
""",
        ),
        sep="\t",
    )

    output_path = tmp_path / "test_with_index.parquet"
    test_data.query("b % 2 == 0").to_parquet(output_path, engine="pyarrow")
    return output_path


def test_parquet_with_index_column(parquet_with_index_column: UPath):
    fields = DatasetParser.for_file(parquet_with_index_column).get_fields()
    assert not any(f.short_name == "__index_level_0__" for f in fields)



import os
from collections.abc import Generator
from datetime import date
from datetime import datetime
from pathlib import Path
from unittest.mock import Mock

import pytest
from dapla_auth_client import AuthClient
from pytest_mock import MockFixture
from pytest_mock import MockType

from dapla_metadata.variable_definitions._generated import vardef_client
from dapla_metadata.variable_definitions._generated.vardef_client.api_client import (
    ApiClient,
)
from dapla_metadata.variable_definitions._generated.vardef_client.configuration import (
    Configuration,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.complete_response import (
    CompleteResponse,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.contact import (
    Contact,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.draft import (
    Draft,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.language_string_type import (
    LanguageStringType,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.owner import (
    Owner,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.patch import (
    Patch,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.update_draft import (
    UpdateDraft,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.validity_period import (
    ValidityPeriod,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.variable_status import (
    VariableStatus,
)
from dapla_metadata.variable_definitions._utils._client import VardefClient
from dapla_metadata.variable_definitions._utils.config import WORKSPACE_DIR
from dapla_metadata.variable_definitions._utils.constants import (
    VARIABLE_DEFINITIONS_DIR,
)
from dapla_metadata.variable_definitions._utils.template_files import (
    create_template_yaml,
)
from dapla_metadata.variable_definitions.variable_definition import VariableDefinition
from tests.utils.constants import VARDEF_EXAMPLE_DEFINITION_ID
from tests.utils.constants import VARDEF_EXAMPLE_INVALID_ID
from tests.utils.microcks_testcontainer import MicrocksContainer
from tests.variable_definitions.constants import OPENAPI_DIR


class CouldNotInstantiateTestContainerError(RuntimeError):
    def __init__(self):
        """To be raised when the TestContainer cannot be instantiated."""
        super().__init__(
            "Could not instantiate the TestContainer. Please check that your local Docker is correctly configured."
        )


@pytest.fixture(scope="session")
def vardef_mock_service() -> Generator[MicrocksContainer | None]:
    with MicrocksContainer() as container:
        container.upload_primary_artifact(
            str(OPENAPI_DIR / "variable-definitions-internal.yml"),
        )
        yield container


@pytest.fixture(scope="session")
def client_configuration(
    vardef_mock_service: MicrocksContainer | None,
) -> Configuration:
    if vardef_mock_service is None:
        raise CouldNotInstantiateTestContainerError
    return vardef_client.Configuration(
        host=vardef_mock_service.get_mock_url(),
        access_token="test_dummy",  # noqa: S106
    )


@pytest.fixture(autouse=True, scope="session")
def _configure_vardef_client(client_configuration):
    VardefClient.set_config(client_configuration)


@pytest.fixture(autouse=True)
def set_temp_workspace(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
    """Fixture which set env WORKSPACE_DIR to tmp path/work."""
    workspace_dir = tmp_path / "work"
    workspace_dir.mkdir(parents=True, exist_ok=True)
    monkeypatch.setenv(WORKSPACE_DIR, str(workspace_dir))
    return workspace_dir


@pytest.fixture
def api_client(client_configuration) -> Generator[ApiClient]:
    with vardef_client.ApiClient(client_configuration) as api_client:
        yield api_client


@pytest.fixture
def language_string_type() -> LanguageStringType:
    return LanguageStringType(nb="test", nn="test", en="test")


@pytest.fixture
def contact(language_string_type) -> Contact:
    return Contact(title=language_string_type, email="me@example.com")


@pytest.fixture
def owner() -> Owner:
    return Owner(team="my_team", groups=["my_team_developers"])


@pytest.fixture
def draft(language_string_type, contact) -> Draft:
    return Draft(
        name=language_string_type,
        short_name="test",
        definition=language_string_type,
        classification_reference="91",
        unit_types=["01"],
        subject_fields=["a", "b"],
        contains_special_categories_of_personal_data=True,
        measurement_type="test",
        valid_from=date(2024, 11, 1),
        external_reference_uri="http://www.example.com",
        comment=language_string_type,
        related_variable_definition_uris=["http://www.example.com"],
        contact=contact,
    )


def sample_variable_definition() -> VariableDefinition:
    return VariableDefinition(
        id=VARDEF_EXAMPLE_DEFINITION_ID,
        patch_id=1,
        name=LanguageStringType(nb="test", nn="test", en="test"),
        short_name="var_test",
        definition=LanguageStringType(nb="test", nn="test", en="test"),
        classification_reference="91",
        unit_types=["01"],
        subject_fields=["a", "b"],
        contains_special_categories_of_personal_data=True,
        variable_status=VariableStatus.PUBLISHED_EXTERNAL,
        measurement_type="test",
        valid_from=date(2024, 11, 1),
        valid_until=None,
        external_reference_uri="http://www.example.com",
        comment=LanguageStringType(nb="test", nn="test", en="test"),
        related_variable_definition_uris=["http://www.example.com"],
        contact=Contact(
            title=LanguageStringType(nb="test", nn="test", en="test"),
            email="me@example.com",
        ),
        owner=Owner(team="my_team", groups=["my_team_developers"]),
        created_at=date(2024, 11, 1),
        created_by="ano@ssb.no",
        last_updated_at=date(2024, 11, 1),
        last_updated_by="ano@ssb.no",
    )


def unknown_variable_definition() -> VariableDefinition:
    unknown = sample_variable_definition()
    unknown.id = VARDEF_EXAMPLE_INVALID_ID
    return unknown


@pytest.fixture
def variable_definition() -> VariableDefinition:
    return sample_variable_definition()


@pytest.fixture
def update_draft(language_string_type, contact, owner) -> UpdateDraft:
    return UpdateDraft(
        name=language_string_type,
        short_name="test",
        definition=language_string_type,
        classification_reference="91",
        unit_types=["a", "b"],
        subject_fields=["a", "b"],
        contains_special_categories_of_personal_data=True,
        variable_status=VariableStatus.PUBLISHED_EXTERNAL,
        measurement_type="test",
        valid_from=date(2024, 11, 1),
        external_reference_uri="http://www.example.com",
        comment=language_string_type,
        related_variable_definition_uris=["http://www.example.com"],
        contact=contact,
        owner=owner,
    )


@pytest.fixture
def validity_period(language_string_type, contact) -> ValidityPeriod:
    return ValidityPeriod(
        name=language_string_type,
        definition=language_string_type,
        classification_reference="91",
        unit_types=["a", "b"],
        subject_fields=["a", "b"],
        contains_special_categories_of_personal_data=True,
        variable_status=VariableStatus.PUBLISHED_EXTERNAL,
        measurement_type="test",
        valid_from=date(2024, 11, 1),
        external_reference_uri="http://www.example.com",
        comment=language_string_type,
        related_variable_definition_uris=["http://www.example.com"],
        contact=contact,
    )


@pytest.fixture
def patch_fixture(language_string_type, contact, owner) -> Patch:
    return Patch(
        name=language_string_type,
        definition=language_string_type,
        classification_reference="91",
        unit_types=["a", "b"],
        subject_fields=["a", "b"],
        contains_special_categories_of_personal_data=True,
        variable_status=VariableStatus.PUBLISHED_EXTERNAL,
        measurement_type="test",
        external_reference_uri="http://www.example.com",
        comment=language_string_type,
        related_variable_definition_uris=["http://www.example.com"],
        contact=contact,
        owner=owner,
    )


def sample_complete_patch_output() -> VariableDefinition:
    return VariableDefinition(
        id=VARDEF_EXAMPLE_DEFINITION_ID,
        patch_id=1,
        name=LanguageStringType(nb="test", nn="test", en="test"),
        short_name="var_test",
        definition=LanguageStringType(nb="test", nn="test", en="test"),
        classification_reference="91",
        unit_types=["01"],
        subject_fields=["a", "b"],
        contains_special_categories_of_personal_data=True,
        variable_status=VariableStatus.PUBLISHED_INTERNAL,
        measurement_type="test",
        valid_from=date(2024, 11, 1),
        valid_until=None,
        external_reference_uri="http://www.example.com",
        comment=LanguageStringType(nb="test", nn="test", en="test"),
        related_variable_definition_uris=["http://www.example.com"],
        contact=Contact(
            title=LanguageStringType(nb="test", nn="test", en="test"),
            email="me@example.com",
        ),
        owner=Owner(team="my_team", groups=["my_team_developers"]),
        created_at=date(2024, 11, 1),
        created_by="ano@ssb.no",
        last_updated_at=date(2024, 11, 1),
        last_updated_by="ano@ssb.no",
    )


@pytest.fixture
def mock_update_variable_definition_by_id():
    return Mock(
        return_value=CompleteResponse(
            id="test-id",
            patch_id=1,
            name=LanguageStringType(en="Name", nb="Navn"),
            short_name="short-name",
            definition=LanguageStringType(en="Definition", nb="Definition"),
            unit_types=["unit-type"],
            subject_fields=["subject-field"],
            contains_special_categories_of_personal_data=False,
            variable_status=VariableStatus.DRAFT,
            valid_from=date(2023, 1, 1),
            owner=Owner(team="Owner", groups=["dapla-felles"]),
            contact=Contact(
                title=LanguageStringType(en="Contact", nb="Contact"),
                email="contact@contact.com",
            ),
            classification_reference="www.newurl.com",
            created_at=datetime(2023, 1, 1, 12, 0),  # noqa: DTZ001
            created_by="created by",
            last_updated_at=datetime(2023, 1, 1, 12, 0),  # noqa: DTZ001
            last_updated_by="updated by",
        ),
    )


@pytest.fixture
def complete_patch_output() -> VariableDefinition:
    return sample_complete_patch_output()


@pytest.fixture
def work_folder_defaults(set_temp_workspace: Path):
    """Fixture that ensures a work folder exists for template with default values."""
    base_path = set_temp_workspace
    file_name = create_template_yaml(
        custom_directory=base_path,
    )

    return base_path / file_name


@pytest.fixture
def work_folder_complete_patch_output(set_temp_workspace: Path):
    """Fixture that ensures a work folder exists for template with saved variable definition values."""
    base_path = set_temp_workspace / VARIABLE_DEFINITIONS_DIR
    file_name = create_template_yaml(
        sample_complete_patch_output(),
        custom_directory=base_path,
    )
    return base_path / file_name


@pytest.fixture
def _delete_workspace_dir_env_var(set_temp_workspace):  # noqa: ARG001
    original_workspace_dir = os.environ.get(WORKSPACE_DIR)

    if WORKSPACE_DIR in os.environ:
        del os.environ[WORKSPACE_DIR]

    yield

    if original_workspace_dir is not None:
        os.environ[WORKSPACE_DIR] = original_workspace_dir
    else:
        pass


@pytest.fixture
def work_folder_variable_definition(set_temp_workspace: Path):
    """Fixture that ensures a work folder exists for template with saved variable definition values."""
    base_path = set_temp_workspace
    file_name = create_template_yaml(
        sample_variable_definition(),
        custom_directory=base_path,
    )
    return base_path / file_name


VARIABLE_DEFINITION_DICT = {
    "name": {
        "en": "Country Background",
        "nb": "Landbakgrunn",
        "nn": "Landbakgrunn",
    },
    "short_name": "new_short_name1",
    "definition": {
        "en": "Country background is the person's own, the mother's or possibly the father's country of birth. Persons without an immigrant background always have Norway as country background. In cases where the parents have different countries of birth the mother's country of birth is chosen. If neither the person nor the parents are born abroad, country background is chosen from the first person born abroad in the order mother's mother, mother's father, father's mother, father's father.",
        "nb": "For personer født i utlandet, er dette (med noen få unntak) eget fødeland. For personer født i Norge er det foreldrenes fødeland. I de tilfeller der foreldrene har ulikt fødeland, er det morens fødeland som blir valgt. Hvis ikke personen selv eller noen av foreldrene er utenlandsfødt, hentes landbakgrunn fra de første utenlandsfødte en treffer på i rekkefølgen mormor, morfar, farmor eller farfar.",
        "nn": "For personar fødd i utlandet, er dette (med nokre få unntak) eige fødeland. For personar fødd i Noreg er det fødelandet til foreldra. I dei tilfella der foreldra har ulikt fødeland, er det fødelandet til mora som blir valt. Viss ikkje personen sjølv eller nokon av foreldra er utenlandsfødt, blir henta landsbakgrunn frå dei første utenlandsfødte ein treffar på i rekkjefølgja mormor, morfar, farmor eller farfar.",
    },
    "classification_reference": "91",
    "unit_types": ["01", "02"],
    "subject_fields": ["he04"],
    "contains_special_categories_of_personal_data": True,
    "measurement_type": None,
    "valid_from": "2003-01-01",
    "external_reference_uri": "https://www.ssb.no/a/metadata/conceptvariable/vardok/1919/nb",
    "comment": {
        "nb": "Fra og med 1.1.2003 ble definisjon endret til også å trekke inn besteforeldrenes fødeland.",
        "nn": "Fra og med 1.1.2003 ble definisjon endret til også å trekke inn besteforeldrenes fødeland.",
        "en": "As of 1 January 2003, the definition was changed to also include the grandparents' country of birth.",
    },
    "related_variable_definition_uris": ["https://example.com/"],
    "contact": {
        "title": {
            "en": "Division for population statistics",
            "nb": "Seksjon for befolkningsstatistikk",
            "nn": "Seksjon for befolkningsstatistikk",
        },
        "email": "s320@ssb.no",
    },
}


@pytest.fixture
def set_workspace_not_dir(tmp_path: Path):
    file_path = tmp_path / "funnyfiles.txt"
    file_path.write_text("This is a text file.")
    return file_path


@pytest.fixture(autouse=True)
def mock_auth_client_fetch_personal_token(
    mocker: MockFixture, fake_labid_jwt: str
) -> MockType:
    """Mock the fetch_personal_token method.

    This method can only run on Dapla Lab so we mock it for all the tests.

    Args:
        mocker (MockFixture): The pytest mocker fixture
        fake_labid_jwt (str): An example JWT token

    Returns:
        MockType: The finally configured mock.
    """
    return mocker.patch.object(
        AuthClient,
        "fetch_personal_token",
        autospec=True,
        return_value=fake_labid_jwt,
    )



from datetime import datetime
from datetime import timezone
from typing import Any

from dapla_metadata.datasets.compatibility._utils import DATADOC_KEY
from dapla_metadata.datasets.compatibility._utils import DATASET_KEY
from dapla_metadata.datasets.compatibility._utils import DOCUMENT_VERSION_KEY
from dapla_metadata.datasets.compatibility._utils import PSEUDONYMIZATION_KEY
from dapla_metadata.datasets.compatibility._utils import VARIABLES_KEY
from dapla_metadata.datasets.compatibility._utils import add_container
from dapla_metadata.datasets.compatibility._utils import cast_to_date_type
from dapla_metadata.datasets.compatibility._utils import convert_datetime_to_date
from dapla_metadata.datasets.compatibility._utils import convert_is_personal_data
from dapla_metadata.datasets.compatibility._utils import copy_pseudonymization_metadata
from dapla_metadata.datasets.compatibility._utils import (
    find_and_update_language_strings,
)
from dapla_metadata.datasets.compatibility._utils import remove_element_from_model


def handle_current_version(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle the current version of the metadata.

    This function returns the supplied metadata unmodified.

    Args:
        supplied_metadata: The metadata for the current version.

    Returns:
        The unmodified supplied metadata.
    """
    return supplied_metadata


def handle_version_6_0_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 6.1.0.

    This function modifies the supplied metadata to accommodate breaking changes
    introduced in version 6.1.0. Specifically, it:
    - Consolidates `use_restriction` and `use_restriction_date` into a list of
      dictionaries under `use_restrictions`.
    - Removes the old `use_restriction` and `use_restriction_date` fields.
    - It also converts `use_restriction_date` from datetime to date.

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The upgraded metadata dictionary.
    """
    dataset = supplied_metadata[DATADOC_KEY][DATASET_KEY]

    use_restriction = dataset.get("use_restriction")
    if use_restriction is not None:
        converted_date = convert_datetime_to_date(dataset.get("use_restriction_date"))
        dataset["use_restrictions"] = [
            {
                "use_restriction_type": use_restriction,
                "use_restriction_date": converted_date,
            }
        ]
    else:
        dataset["use_restrictions"] = []

    for field in ("use_restriction", "use_restriction_date"):
        remove_element_from_model(dataset, field)

    return supplied_metadata


def handle_version_5_0_1(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 6.0.0.

    This function modifies the supplied metadata to accommodate breaking changes
    introduced in version 6.0.0. Specifically, it:
    - Moves the following fields from dataset to variable level:
        - contains_personal_data (becomes is_personal_data)
        - unit_type
        - data_source
        - temporality_type

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The updated metadata dictionary.
    """
    fields = [
        ("contains_personal_data", "is_personal_data"),
        ("unit_type", "unit_type"),
        ("data_source", "data_source"),
        ("temporality_type", "temporality_type"),
    ]

    dataset: dict[str, Any] = supplied_metadata[DATADOC_KEY][DATASET_KEY]
    variables: list[dict[str, Any]] = supplied_metadata[DATADOC_KEY][VARIABLES_KEY]

    for f in fields:
        dataset_level_field_value = dataset.pop(f[0], None)
        for v in variables:
            if v.get(f[1]) is None:
                # Don't override any set values
                v[f[1]] = dataset_level_field_value

    return supplied_metadata


def handle_version_4_0_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 5.0.1.

    This function modifies the supplied metadata to accommodate breaking changes
    introduced in version 5.0.1. Specifically, it:
    - Copies pseudonymization metadata if pseudonymization is enabled.
    - Converts the 'is_personal_data' fields to be a bool.
    - All 'pseudonymization' from the container is removed.
    - It also updates the container version to 1.0.0 from 0.0.1

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The updated metadata dictionary.
    """
    if supplied_metadata.get(PSEUDONYMIZATION_KEY):
        copy_pseudonymization_metadata(supplied_metadata)

    convert_is_personal_data(supplied_metadata)

    remove_element_from_model(supplied_metadata, PSEUDONYMIZATION_KEY)
    supplied_metadata[DOCUMENT_VERSION_KEY] = "1.0.0"
    return supplied_metadata


def handle_version_3_3_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 3.3.0.

    This function modifies the supplied metadata to accommodate breaking changes
    introduced in version 4.0.0. Specifically, it removes the
    'direct_person_identifying' field from each variable in 'datadoc.variables'.

    Version 4.0.0 used an enum for is_personal_data, however this was changed to a bool again for version 5.0.1.
    We skip setting the enum here and just keep the value it has.

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The updated metadata dictionary.
    """
    for variable in supplied_metadata[DATADOC_KEY][VARIABLES_KEY]:
        variable["is_personal_data"] = variable["direct_person_identifying"]
        remove_element_from_model(variable, "direct_person_identifying")

    return supplied_metadata


def handle_version_3_2_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 3.2.0.

    This function modifies the supplied metadata to accommodate breaking
    changes introduced in version 3.3.0. Specifically, it updates the
    'contains_data_from' and 'contains_data_until' fields in both the 'dataset'
    and 'variables' sections of the supplied metadata dictionary to ensure they
    are stored as date strings.

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The updated metadata dictionary.
    """
    fields = ["contains_data_from", "contains_data_until"]
    for field in fields:
        supplied_metadata[DATADOC_KEY][DATASET_KEY][field] = cast_to_date_type(
            supplied_metadata[DATADOC_KEY][DATASET_KEY].get(field, None),
        )
        for v in supplied_metadata[DATADOC_KEY][VARIABLES_KEY]:
            v[field] = cast_to_date_type(v.get(field, None))

    return supplied_metadata


def handle_version_3_1_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 3.1.0.

    This function modifies the supplied metadata to accommodate breaking
    changes introduced in version 3.2.0. Specifically, it updates the
    'data_source' field in both the 'dataset' and 'variables' sections of the
    supplied metadata dictionary by converting value to string.

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The updated metadata dictionary.
    """
    data = supplied_metadata[DATADOC_KEY][DATASET_KEY]["data_source"]

    if data is not None:
        supplied_metadata[DATADOC_KEY][DATASET_KEY]["data_source"] = str(
            data[0]["languageText"],
        )

    for i in range(len(supplied_metadata[DATADOC_KEY][VARIABLES_KEY])):
        data = supplied_metadata[DATADOC_KEY][VARIABLES_KEY][i]["data_source"]
        if data is not None:
            supplied_metadata[DATADOC_KEY][VARIABLES_KEY][i]["data_source"] = str(
                data[0]["languageText"],
            )

    return supplied_metadata


def handle_version_2_2_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 2.2.0.

    This function modifies the supplied metadata to accommodate breaking changes
    introduced in version 3.1.0. Specifically, it updates the 'subject_field' in
    the 'dataset' section of the supplied metadata dictionary by converting it to
    a string. It also removes the 'register_uri' field from the 'dataset'.
    Additionally, it removes 'sentinel_value_uri' from each variable,
    sets 'special_value' and 'custom_type' fields to None, and updates
    language strings in the 'variables' and 'dataset' sections.

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The updated metadata dictionary.
    """
    if supplied_metadata[DATADOC_KEY][DATASET_KEY]["subject_field"] is not None:
        data = supplied_metadata[DATADOC_KEY][DATASET_KEY]["subject_field"]
        supplied_metadata[DATADOC_KEY][DATASET_KEY]["subject_field"] = str(
            data["nb"] or data["nn"] or data["en"],
        )

    remove_element_from_model(
        supplied_metadata[DATADOC_KEY][DATASET_KEY], "register_uri"
    )

    for i in range(len(supplied_metadata[DATADOC_KEY][VARIABLES_KEY])):
        remove_element_from_model(
            supplied_metadata[DATADOC_KEY][VARIABLES_KEY][i],
            "sentinel_value_uri",
        )
        supplied_metadata[DATADOC_KEY][VARIABLES_KEY][i]["special_value"] = None
        supplied_metadata[DATADOC_KEY][VARIABLES_KEY][i]["custom_type"] = None
        supplied_metadata[DATADOC_KEY][VARIABLES_KEY][i] = (
            find_and_update_language_strings(
                supplied_metadata[DATADOC_KEY][VARIABLES_KEY][i],
            )
        )
    supplied_metadata[DATADOC_KEY][DATASET_KEY]["custom_type"] = None
    supplied_metadata[DATADOC_KEY][DATASET_KEY] = find_and_update_language_strings(
        supplied_metadata[DATADOC_KEY][DATASET_KEY],
    )
    return supplied_metadata


def handle_version_2_1_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 2.1.0.

    This function modifies the supplied metadata to accommodate breaking changes
    introduced in version 2.2.0. Specifically, it updates the 'owner' field in
    the 'dataset' section of the supplied metadata dictionary by converting it
    from a LanguageStringType to a string.

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The updated metadata dictionary.
    """
    data = supplied_metadata[DATASET_KEY]["owner"]
    supplied_metadata[DATASET_KEY]["owner"] = str(
        data["nb"] or data["nn"] or data["en"]
    )
    return add_container(supplied_metadata)


def handle_version_1_0_0(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 1.0.0.

    This function modifies the supplied metadata to accommodate breaking changes
    introduced in version 2.1.0. Specifically, it updates the date fields
    'metadata_created_date' and 'metadata_last_updated_date' to ISO 8601 format
    with UTC timezone. It also converts the 'data_source' field from a string to a
    dictionary with language keys if necessary and removes the 'data_source_path'
    field.

    Args:
        supplied_metadata: The metadata dictionary to be updated.

    Returns:
        The updated metadata dictionary.

    """
    datetime_fields = [("metadata_created_date"), ("metadata_last_updated_date")]
    for field in datetime_fields:
        if supplied_metadata[DATASET_KEY][field]:
            supplied_metadata[DATASET_KEY][field] = datetime.isoformat(
                datetime.fromisoformat(
                    supplied_metadata[DATASET_KEY][field]
                ).astimezone(
                    tz=timezone.utc,
                ),
                timespec="seconds",
            )
    if isinstance(supplied_metadata[DATASET_KEY]["data_source"], str):
        supplied_metadata[DATASET_KEY]["data_source"] = {
            "en": supplied_metadata[DATASET_KEY]["data_source"],
            "nn": "",
            "nb": "",
        }

    remove_element_from_model(supplied_metadata[DATASET_KEY], "data_source_path")

    return supplied_metadata


def handle_version_0_1_1(supplied_metadata: dict[str, Any]) -> dict[str, Any]:
    """Handle breaking changes for version 0.1.1.

    This function modifies the supplied metadata to accommodate breaking changes
    introduced in version 1.0.0. Specifically, it renames certain keys within the
    `dataset` and `variables` sections, and replaces empty string values with
    `None` for `dataset` keys.

    Args:
        supplied_metadata: The metadata dictionary that needs to be updated.

    Returns:
        The updated metadata dictionary.

    References:
        PR ref: https://github.com/statisticsnorway/ssb-datadoc-model/pull/4
    """
    key_renaming = [
        ("metadata_created_date", "created_date"),
        ("metadata_created_by", "created_by"),
        ("metadata_last_updated_date", "last_updated_date"),
        ("metadata_last_updated_by", "last_updated_by"),
    ]
    for new_key, old_key in key_renaming:
        supplied_metadata[DATASET_KEY][new_key] = supplied_metadata[DATASET_KEY].pop(
            old_key,
        )
    # Replace empty strings with None, empty strings are not valid for LanguageStrings values
    supplied_metadata[DATASET_KEY] = {
        k: None if v == "" else v for k, v in supplied_metadata[DATASET_KEY].items()
    }

    key_renaming = [("data_type", "datatype")]

    for i in range(len(supplied_metadata[VARIABLES_KEY])):
        for new_key, old_key in key_renaming:
            supplied_metadata[VARIABLES_KEY][i][new_key] = supplied_metadata[
                VARIABLES_KEY
            ][i].pop(
                old_key,
            )

    return supplied_metadata



import contextlib
from pathlib import Path
from unittest.mock import MagicMock
from unittest.mock import patch

import pytest
from pydantic import ValidationError

from dapla_metadata.variable_definitions._generated.vardef_client.models.patch import (
    Patch,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.update_draft import (
    UpdateDraft,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.validity_period import (
    ValidityPeriod,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.variable_status import (
    VariableStatus,
)
from dapla_metadata.variable_definitions._utils._client import VardefClient
from dapla_metadata.variable_definitions._utils.constants import (
    PUBLISHING_BLOCKED_ERROR_MESSAGE,
)
from dapla_metadata.variable_definitions._utils.constants import VARDEF_PROD_URL
from dapla_metadata.variable_definitions._utils.constants import VARDEF_TEST_URL
from dapla_metadata.variable_definitions.exceptions import PublishingBlockedError
from dapla_metadata.variable_definitions.exceptions import VardefFileError
from dapla_metadata.variable_definitions.vardef import Vardef
from dapla_metadata.variable_definitions.variable_definition import (
    IDENTICAL_PATCH_ERROR_MESSAGE,
)
from dapla_metadata.variable_definitions.variable_definition import VariableDefinition
from tests.utils.constants import VARDEF_EXAMPLE_DATE
from tests.utils.constants import VARDEF_EXAMPLE_DEFINITION_ID
from tests.variable_definitions.constants import VARIABLE_DEFINITION_EDITING_FILES_DIR
from tests.variable_definitions.test_vardef import PATCH_ID


def test_list_patches():
    landbak = Vardef.get_variable_definition_by_id(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
    )
    assert isinstance(landbak.list_patches()[0], VariableDefinition)


def test_get_patch():
    landbak = Vardef.get_variable_definition_by_id(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
    )
    assert isinstance(landbak.get_patch(1), VariableDefinition)


def test_list_validity_periods():
    landbak = Vardef.get_variable_definition_by_id(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
    )
    assert isinstance(landbak.list_validity_periods()[0], VariableDefinition)


def test_update_draft(
    update_draft: UpdateDraft,
):
    my_draft = Vardef.get_variable_definition_by_id(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
    )
    assert isinstance(my_draft.update_draft(update_draft), VariableDefinition)


def test_update_draft_check_model_change(
    update_draft: UpdateDraft,
    mock_update_variable_definition_by_id,
):
    my_draft = Vardef.get_variable_definition_by_id(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
    )
    with patch(
        "dapla_metadata.variable_definitions._generated.vardef_client.api.draft_variable_definitions_api.DraftVariableDefinitionsApi.update_variable_definition_by_id",
        mock_update_variable_definition_by_id,
    ):
        my_draft.update_draft(update_draft=update_draft)
        assert my_draft.classification_reference == "www.newurl.com"
        mock_update_variable_definition_by_id.assert_called_once()


def test_delete_draft():
    my_draft = Vardef.get_variable_definition_by_id(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
    )
    assert my_draft.id is not None
    result = my_draft.delete_draft()
    assert f"Variable {VARDEF_EXAMPLE_DEFINITION_ID} safely deleted" in result


def test_create_patch(
    patch_fixture: Patch,
    variable_definition: VariableDefinition,
):
    my_variable = variable_definition
    assert my_variable.patch_id == 1
    created_patch = my_variable.create_patch(
        patch=patch_fixture,
        valid_from=VARDEF_EXAMPLE_DATE,
    )
    assert isinstance(
        created_patch,
        VariableDefinition,
    )
    assert created_patch.patch_id == PATCH_ID


def test_create_patch_check_model_change(
    patch_fixture: Patch,
    mock_update_variable_definition_by_id,
):
    my_draft = Vardef.get_variable_definition_by_id(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
    )
    with patch(
        "dapla_metadata.variable_definitions._generated.vardef_client.api.patches_api.PatchesApi.create_patch",
        mock_update_variable_definition_by_id,
    ):
        my_draft.create_patch(patch=patch_fixture)
        assert my_draft.classification_reference == "www.newurl.com"
        mock_update_variable_definition_by_id.assert_called_once()


def test_create_validity_period(
    validity_period: ValidityPeriod,
    variable_definition: VariableDefinition,
):
    my_variable = variable_definition
    assert isinstance(
        my_variable.create_validity_period(validity_period),
        VariableDefinition,
    )


def test_create_validity_period_check_model_change(
    validity_period: ValidityPeriod,
    mock_update_variable_definition_by_id,
):
    my_draft = Vardef.get_variable_definition_by_id(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
    )
    with patch(
        "dapla_metadata.variable_definitions._generated.vardef_client.api.validity_periods_api.ValidityPeriodsApi.create_validity_period",
        mock_update_variable_definition_by_id,
    ):
        my_draft.create_validity_period(validity_period=validity_period)
        assert my_draft.classification_reference == "www.newurl.com"
        mock_update_variable_definition_by_id.assert_called_once()


def test_update_draft_from_file():
    my_draft = Vardef.get_variable_definition_by_id(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
    ).to_file()
    assert my_draft.get_file_path().exists()
    assert isinstance(my_draft.update_draft_from_file(), VariableDefinition)


def test_update_draft_from_file_no_known_file():
    with pytest.raises(FileNotFoundError, match="Could not deduce a path to the file"):
        Vardef.get_variable_definition_by_id(
            VARDEF_EXAMPLE_DEFINITION_ID,
        ).update_draft_from_file()


def test_update_draft_from_file_file_non_existent():
    with pytest.raises(FileNotFoundError):
        Vardef.get_variable_definition_by_id(
            VARDEF_EXAMPLE_DEFINITION_ID,
        ).update_draft_from_file("my_cool_file.yaml")


def test_update_draft_from_file_invalid_yaml(tmp_path: Path):
    invalid = tmp_path / "invalid_yaml.yaml"
    invalid.write_text(
        """--- invalid yaml ---
not allowed:
:why
""",
    )
    with pytest.raises(VardefFileError):
        Vardef.get_variable_definition_by_id(
            VARDEF_EXAMPLE_DEFINITION_ID,
        ).update_draft_from_file(
            invalid,
        )


def test_update_draft_from_file_validation_fails(tmp_path: Path):
    invalid = tmp_path / "fails_validation.yaml"
    invalid.write_text(
        """fails validation""",
    )
    with pytest.raises(ValidationError):
        Vardef.get_variable_definition_by_id(
            VARDEF_EXAMPLE_DEFINITION_ID,
        ).update_draft_from_file(
            invalid,
        )


def test_update_draft_from_file_specify_file():
    assert (
        Vardef.get_variable_definition_by_id(
            VARDEF_EXAMPLE_DEFINITION_ID,
        )
        .update_draft_from_file(
            VARIABLE_DEFINITION_EDITING_FILES_DIR / "classification_reference.yaml",
        )
        .short_name
        == "landbak"
    )


def test_create_patch_from_file(variable_definition: VariableDefinition):
    mock_api_response = MagicMock(name="VariableDefinition")
    mock_api_response.classification_reference = "702"
    with patch.object(
        VariableDefinition,
        "create_patch",
        return_value=mock_api_response,
    ) as mock_create_patch:
        my_patch = variable_definition
        result = my_patch.create_patch_from_file(
            file_path=VARIABLE_DEFINITION_EDITING_FILES_DIR
            / "classification_reference.yaml",
        )

    assert my_patch.classification_reference != result.classification_reference
    mock_create_patch.assert_called_once()


def test_create_patch_change_short_name(variable_definition: VariableDefinition):
    mock_api_response = MagicMock(name="VariableDefinition")
    mock_api_response.classification_reference = "702"
    with (
        patch.object(
            VariableDefinition,
            "create_patch",
            return_value=mock_api_response,
        ) as mock_create_patch,
        pytest.raises(ValueError, match=IDENTICAL_PATCH_ERROR_MESSAGE),
    ):
        variable_definition.create_patch_from_file(
            file_path=VARIABLE_DEFINITION_EDITING_FILES_DIR
            / "variable_definition_var_test_wypvb3wd_2025-09-18T14-19-28.yaml",
        )

    mock_create_patch.assert_not_called()


def test_from_model(
    variable_definition: VariableDefinition,
):
    """This covers a bug that was reported in.

    It was caused by an incorrect implementation in the from_model
    method. The only thing we're testing is that no exceptions are
    raised in this flow.
    """
    mock_api_response = MagicMock(name="VariableDefinition")
    mock_api_response.classification_reference = "702"
    with patch.object(
        VariableDefinition,
        "create_patch",
        return_value=mock_api_response,
    ) as mock_create_patch:
        my_patch = VariableDefinition.from_model(variable_definition)
        my_patch.create_patch_from_file(
            file_path=VARIABLE_DEFINITION_EDITING_FILES_DIR
            / "variable_definition_landbak_wypvb3wd_2025-05-02T10-06-20.yaml",
        )
        mock_create_patch.assert_called_once()


def test_create_patch_from_file_path_not_set(variable_definition: VariableDefinition):
    my_patch = variable_definition
    my_patch.set_file_path(file_path=None)
    with pytest.raises(FileNotFoundError, match="Could not deduce a path to the file"):
        my_patch.create_patch_from_file()


def test_create_patch_from_file_path_non_existing(
    variable_definition: VariableDefinition,
):
    my_patch = variable_definition
    with pytest.raises(FileNotFoundError):
        my_patch.create_patch_from_file("some_file.yaml")


def test_create_validity_period_from_file(variable_definition: VariableDefinition):
    mock_api_response = MagicMock(name="VariableDefinition")
    mock_api_response.valid_from = "2025-01-01"
    mock_api_response.definition = {
        "nb": "Ny definition",
        "nn": "Ny definition",
        "en": "New definition",
    }

    with patch.object(
        VariableDefinition,
        "create_validity_period",
        return_value=mock_api_response,
    ) as mock_create_validity_period:
        prev_variable_definition = variable_definition
        result = prev_variable_definition.create_validity_period_from_file(
            file_path=VARIABLE_DEFINITION_EDITING_FILES_DIR
            / "date_and_definition.yaml",
        )
    assert result.definition != prev_variable_definition.definition
    assert result.valid_from != prev_variable_definition.valid_from
    mock_create_validity_period.assert_called_once()


def test_create_draft_with_special_characters():
    mock_api_response = MagicMock(name="VariableDefinition")
    mock_api_response.definition = {
        "nb": "Varigheten av en konflikt måles i antall tapte arbeidsdager, dvs. virkedager per uke for den aktuelle gruppen i konflikt, ikke kalenderdager. Det kan innebære mange spesialtegn som : og ;",
        "nn": "",
        "en": "Duration of an industrial disputes contains in number of work-days lost, i.e. working-days a week to the actual group in work stoppages, not calendar days.",
    }

    with patch.object(
        Vardef,
        "create_draft_from_file",
        return_value=mock_api_response,
    ) as mock_create_draft:
        result = Vardef.create_draft_from_file(
            file_path=VARIABLE_DEFINITION_EDITING_FILES_DIR
            / "variable_definition_landbak_wypvb3wd_2025-05-02T10-06-20.yaml",
        )
    assert ":" in result.definition["nb"]
    assert result.definition["nn"] == ""
    mock_create_draft.assert_called_once()


def test_update_draft_with_special_characters(variable_definition: VariableDefinition):
    mock_api_response = MagicMock(name="VariableDefinition")
    mock_api_response.definition = {
        "nb": "Ny definition med :",
        "nn": "Ny definition med %&",
        "en": "New definition",
    }

    with patch.object(
        VariableDefinition,
        "update_draft",
        return_value=mock_api_response,
    ) as mock_update_draft:
        prev_variable_definition = variable_definition
        result = prev_variable_definition.update_draft_from_file(
            file_path=VARIABLE_DEFINITION_EDITING_FILES_DIR
            / "variable_definition_landbak_wypvb3wd_2025-05-02T10-06-20.yaml",
        )
    assert result.definition != prev_variable_definition.definition
    assert ":" in result.definition["nb"]
    assert "&" in result.definition["nn"]
    mock_update_draft.assert_called_once()


def test_create_validity_period_from_file_path_not_set(
    variable_definition: VariableDefinition,
):
    variable_definition.set_file_path(file_path=None)
    with pytest.raises(FileNotFoundError, match="Could not deduce a path to the file"):
        variable_definition.create_patch_from_file()


def test_create_validity_period_from_file_path_non_existing(
    variable_definition: VariableDefinition,
):
    with pytest.raises(FileNotFoundError):
        variable_definition.create_validity_period_from_file("some_file.yaml")


@patch.object(VariableDefinition, "update_draft")
@patch.object(VariableDefinition, "create_patch")
@pytest.mark.parametrize(("initial_status"), list(VariableStatus))
@pytest.mark.parametrize(
    ("method_name", "expected_status"),
    [
        ("publish_internal", VariableStatus.PUBLISHED_INTERNAL),
        ("publish_external", VariableStatus.PUBLISHED_EXTERNAL),
    ],
)
def test_publish_methods(
    mock_create_patch: MagicMock,
    mock_update_draft: MagicMock,
    variable_definition: VariableDefinition,
    initial_status: VariableStatus,
    method_name: str,
    expected_status: VariableStatus,
):
    variable_definition.variable_status = initial_status
    method_to_call = getattr(variable_definition, method_name)
    if initial_status is VariableStatus.PUBLISHED_EXTERNAL:
        # It doesn't make sense to publish other statuses internally
        with pytest.raises(ValueError, match="That won't work here."):
            method_to_call()
    elif method_name == "publish_internal":
        if initial_status is VariableStatus.DRAFT:
            # Normal publishing flow
            method_to_call()
            mock_update_draft.assert_called_once_with(
                UpdateDraft(variable_status=expected_status),
            )
            mock_create_patch.assert_not_called()
        else:
            # It doesn't make sense to publish other statuses internally
            with pytest.raises(ValueError, match="That won't work here."):
                method_to_call()
    else:
        method_to_call()
        mock_update_draft.assert_not_called()
        mock_create_patch.assert_called_once_with(
            Patch(variable_status=VariableStatus.PUBLISHED_EXTERNAL),
        )


def test_str(variable_definition):
    assert (
        str(variable_definition)
        == """id: "wypvb3wd"
patch_id: 1
name:
    nb: |-
        test
    nn: |-
        test
    en: |-
        test
short_name: "var_test"
definition:
    nb: |-
        test
    nn: |-
        test
    en: |-
        test
classification_reference: "91"
unit_types:
    - "01"
subject_fields:
    - "a"
    - "b"
contains_special_categories_of_personal_data: true
variable_status: PUBLISHED_EXTERNAL
measurement_type: "test"
valid_from: '2024-11-01'
valid_until:
external_reference_uri: "http://www.example.com"
comment:
    nb: |-
        test
    nn: |-
        test
    en: |-
        test
related_variable_definition_uris:
    - "http://www.example.com"
owner:
    team: "my_team"
    groups:
        - "my_team_developers"
contact:
    title:
        nb: |-
            test
        nn: |-
            test
        en: |-
            test
    email: me@example.com
created_at: '2024-11-01T00:00:00'
created_by: "ano@ssb.no"
last_updated_at: '2024-11-01T00:00:00'
last_updated_by: "ano@ssb.no"
"""
    )


@pytest.mark.parametrize("method_name", ["publish_internal", "publish_external"])
@pytest.mark.parametrize(
    ("mocked_host", "should_raise"),
    [
        ("https://metadata.intern.ssb.no", True),
        ("https://metadata.intern.test.ssb.no", False),
    ],
)
@pytest.mark.parametrize(("initial_status"), list(VariableStatus))
@patch.object(VariableDefinition, "update_draft")
@patch.object(VariableDefinition, "create_patch")
@patch("dapla_metadata.variable_definitions._utils._client.VardefClient.get_config")
def test_block_publish_methods(
    mock_get_config: MagicMock,
    mock_update_draft: MagicMock,
    mock_create_patch: MagicMock,
    method_name: str,
    mocked_host: str,
    should_raise: bool,
    variable_definition: VariableDefinition,
    initial_status: VariableStatus,
):
    mock_config = MagicMock()
    mock_config.host = mocked_host
    mock_get_config.return_value = mock_config

    client = VardefClient()
    config = client.get_config()
    assert config is not None
    assert config.host == mocked_host

    variable_definition.variable_status = initial_status

    method = getattr(variable_definition, method_name)
    if should_raise:
        with pytest.raises(
            PublishingBlockedError, match=PUBLISHING_BLOCKED_ERROR_MESSAGE
        ):
            method()
        mock_update_draft.assert_not_called()
        mock_create_patch.assert_not_called()

    else:
        try:
            method()
            assert mock_update_draft.called or mock_create_patch.called
        except ValueError:
            # Ignore other exceptions for publishing
            contextlib.suppress(ValueError)


@pytest.mark.parametrize(
    ("mocked_host", "update_status", "should_raise"),
    [
        (
            VARDEF_PROD_URL,
            VariableStatus.PUBLISHED_INTERNAL.value,
            True,
        ),
        (
            VARDEF_PROD_URL,
            VariableStatus.PUBLISHED_EXTERNAL.value,
            True,
        ),
        (VARDEF_PROD_URL, VariableStatus.DRAFT.value, False),
        (VARDEF_TEST_URL, VariableStatus.DRAFT.value, False),
        (
            VARDEF_TEST_URL,
            VariableStatus.PUBLISHED_INTERNAL.value,
            False,
        ),
        (
            VARDEF_TEST_URL,
            VariableStatus.PUBLISHED_EXTERNAL.value,
            False,
        ),
    ],
)
@patch("dapla_metadata.variable_definitions._utils._client.VardefClient.get_config")
def test_block_update_variable_status(
    mock_get_config: MagicMock,
    mocked_host: str,
    update_status,
    should_raise: bool,
    variable_definition: VariableDefinition,
):
    mock_config = MagicMock()
    mock_config.host = mocked_host
    mock_get_config.return_value = mock_config

    client = VardefClient()
    config = client.get_config()
    assert config is not None
    assert config.host == mocked_host

    variable_definition.variable_status = VariableStatus.DRAFT

    with patch.object(
        VariableDefinition, "update_draft", wraps=variable_definition.update_draft
    ):
        update = UpdateDraft(variable_status=update_status)

        if should_raise:
            with pytest.raises(
                PublishingBlockedError, match=PUBLISHING_BLOCKED_ERROR_MESSAGE
            ):
                variable_definition.update_draft(update)
        else:
            try:
                variable_definition.update_draft(update)
            except ValueError:
                # Ignore other exceptions for publishing
                contextlib.suppress(ValueError)



"""Tests for the DataDocMetadata class."""

from __future__ import annotations

import json
import pathlib
import shutil
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import MagicMock
from unittest.mock import patch
from uuid import UUID

import arrow
import pytest
from datadoc_model.all_optional.model import Assessment
from datadoc_model.all_optional.model import DatadocMetadata
from datadoc_model.all_optional.model import Dataset
from datadoc_model.all_optional.model import DataSetState
from datadoc_model.all_optional.model import DataSetStatus
from datadoc_model.all_optional.model import DataType
from datadoc_model.all_optional.model import Variable
from datadoc_model.all_optional.model import VariableRole
from pydantic import ValidationError

from dapla_metadata.dapla.user_info import TestUserInfo
from dapla_metadata.datasets.core import Datadoc
from dapla_metadata.datasets.statistic_subject_mapping import StatisticSubjectMapping
from dapla_metadata.datasets.utility.constants import (
    DATASET_FIELDS_FROM_EXISTING_METADATA,
)
from tests.datasets.constants import DATADOC_METADATA_MODULE_CORE
from tests.datasets.constants import TEST_DATASETS_DIRECTORY
from tests.datasets.constants import TEST_EXISTING_METADATA_DIRECTORY
from tests.datasets.constants import TEST_EXISTING_METADATA_FILE_NAME
from tests.datasets.constants import TEST_EXISTING_METADATA_NAMING_STANDARD_FILEPATH
from tests.datasets.constants import TEST_EXISTING_METADATA_WITH_VALID_ID_DIRECTORY
from tests.datasets.constants import TEST_NAMING_STANDARD_COMPATIBLE_DATASET
from tests.datasets.constants import TEST_PARQUET_FILEPATH
from tests.datasets.constants import TEST_PROCESSED_DATA_POPULATION_DIRECTORY
from tests.datasets.constants import TEST_RESOURCES_DIRECTORY

if TYPE_CHECKING:
    from collections.abc import Generator
    from datetime import datetime

    from upath import UPath
    from upath.types import ReadablePathLike


@pytest.fixture
def generate_periodic_file(
    existing_data_path: UPath,
    insert_string: str,
) -> Generator[UPath, None, None]:
    file_name = existing_data_path.name
    insert_pos = file_name.find("_v1")
    new_file_name = file_name[:insert_pos] + insert_string + file_name[insert_pos:]
    new_path = TEST_RESOURCES_DIRECTORY / new_file_name
    shutil.copy(str(existing_data_path), str(new_path))
    yield new_path
    if new_path.exists():
        new_path.unlink()


@pytest.mark.usefixtures("existing_metadata_file")
def test_existing_metadata_file(
    metadata: Datadoc,
):
    root = getattr(metadata.dataset.name, "root", [])
    if root:
        assert root[0].languageText == "successfully_read_existing_file"
    else:
        msg = "Root is none"
        raise AssertionError(msg)


def test_metadata_document_percent_complete(metadata: Datadoc):
    dataset = Dataset(dataset_state=DataSetState.OUTPUT_DATA)
    variable_1 = Variable(data_type=DataType.BOOLEAN)
    variable_2 = Variable(data_type=DataType.INTEGER)
    document = DatadocMetadata(
        percentage_complete=0,
        dataset=dataset,
        variables=[variable_1, variable_2],
    )
    metadata.dataset = document.dataset  # type: ignore [assignment]
    metadata.variables = document.variables  # type: ignore [assignment]

    assert metadata.percent_complete == 13


def test_write_metadata_document(
    dummy_timestamp: datetime,
    metadata: Datadoc,
    tmp_path: pathlib.Path,
):
    metadata.dataset.metadata_created_date = dummy_timestamp
    metadata.write_metadata_document()
    written_document = tmp_path / TEST_EXISTING_METADATA_FILE_NAME
    assert Path.exists(written_document)
    assert metadata.dataset.metadata_created_date == dummy_timestamp
    assert (
        metadata.dataset.metadata_created_by == TestUserInfo.PLACEHOLDER_EMAIL_ADDRESS
    )
    assert metadata.dataset.metadata_last_updated_date == dummy_timestamp
    assert (
        metadata.dataset.metadata_last_updated_by
        == TestUserInfo.PLACEHOLDER_EMAIL_ADDRESS
    )

    with Path.open(written_document) as f:
        written_metadata = json.loads(f.read())
        datadoc_metadata = written_metadata["datadoc"]["dataset"]

    assert (
        # Use our pydantic model to read in the datetime string so we get the correct format
        Dataset(
            metadata_created_date=datadoc_metadata["metadata_created_date"],
        ).metadata_created_date
        == dummy_timestamp
    )
    assert (
        datadoc_metadata["metadata_created_by"]
        == TestUserInfo.PLACEHOLDER_EMAIL_ADDRESS
    )
    assert (
        # Use our pydantic model to read in the datetime string so we get the correct format
        Dataset(
            metadata_last_updated_date=datadoc_metadata["metadata_last_updated_date"],
        ).metadata_last_updated_date
        == dummy_timestamp
    )
    assert (
        datadoc_metadata["metadata_last_updated_by"]
        == TestUserInfo.PLACEHOLDER_EMAIL_ADDRESS
    )


@pytest.mark.usefixtures("existing_metadata_file")
@patch(
    DATADOC_METADATA_MODULE_CORE + ".user_info.get_user_info_for_current_platform",
    return_value=TestUserInfo(),
)
def test_write_metadata_document_existing_document(
    _mock_user_info: MagicMock,  # noqa: PT019 it's a patch, not a fixture
    dummy_timestamp: datetime,
    metadata: Datadoc,
):
    original_created_date = metadata.dataset.metadata_created_date
    original_created_by = metadata.dataset.metadata_created_by
    metadata.write_metadata_document()
    assert metadata.dataset.metadata_created_by == original_created_by
    assert metadata.dataset.metadata_created_date == original_created_date
    assert (
        metadata.dataset.metadata_last_updated_by
        == TestUserInfo.PLACEHOLDER_EMAIL_ADDRESS
    )
    assert metadata.dataset.metadata_last_updated_date == dummy_timestamp


def test_metadata_id(metadata: Datadoc):
    assert isinstance(metadata.dataset.id, UUID)


@pytest.mark.parametrize(
    "existing_metadata_path",
    [TEST_EXISTING_METADATA_DIRECTORY / "invalid_id_field"],
)
def test_existing_metadata_none_id(
    existing_metadata_file: Path,
    metadata: Datadoc,
):
    with existing_metadata_file.open() as f:
        pre_open_id: None = json.load(f)["datadoc"]["dataset"]["id"]
    assert pre_open_id is None
    assert isinstance(metadata.dataset.id, UUID)
    metadata.write_metadata_document()
    with existing_metadata_file.open() as f:
        post_write_id = json.load(f)["datadoc"]["dataset"]["id"]
    assert post_write_id == str(metadata.dataset.id)


@pytest.mark.parametrize(
    "existing_metadata_path",
    [TEST_EXISTING_METADATA_DIRECTORY / "valid_id_field"],
)
def test_existing_metadata_valid_id(
    existing_metadata_file: Path,
    metadata: Datadoc,
):
    pre_open_id = ""
    post_write_id = ""
    with existing_metadata_file.open() as f:
        pre_open_id = json.load(f)["datadoc"]["dataset"]["id"]
    assert pre_open_id is not None
    assert isinstance(metadata.dataset.id, UUID)
    assert str(metadata.dataset.id) == pre_open_id
    metadata.write_metadata_document()
    with existing_metadata_file.open() as f:
        post_write_id = json.load(f)["datadoc"]["dataset"]["id"]
    assert post_write_id == pre_open_id


@pytest.mark.parametrize(
    "metadata_document",
    [TEST_EXISTING_METADATA_WITH_VALID_ID_DIRECTORY / TEST_EXISTING_METADATA_FILE_NAME],
)
@pytest.mark.usefixtures("_mock_timestamp", "_mock_user_info")
def test_validate_required_fields_incomplete_metadata(
    metadata_document: Path,
    subject_mapping_fake_statistical_structure: StatisticSubjectMapping,
):
    with pytest.raises(ValidationError):
        Datadoc(
            metadata_document_path=str(metadata_document),
            statistic_subject_mapping=subject_mapping_fake_statistical_structure,
            validate_required_fields_on_existing_metadata=True,
        )


@pytest.mark.parametrize(
    "metadata_document",
    [TEST_EXISTING_METADATA_NAMING_STANDARD_FILEPATH],
)
@pytest.mark.usefixtures("_mock_timestamp", "_mock_user_info")
def test_validate_required_fields_complete_metadata(
    metadata_document: Path,
    subject_mapping_fake_statistical_structure: StatisticSubjectMapping,
):
    Datadoc(
        metadata_document_path=str(metadata_document),
        statistic_subject_mapping=subject_mapping_fake_statistical_structure,
        validate_required_fields_on_existing_metadata=True,
    )


def test_dataset_short_name(metadata: Datadoc):
    assert metadata.dataset.short_name == "person_data"


def test_dataset_file_path(metadata: Datadoc):
    assert metadata.dataset.file_path == str(metadata.dataset_path)


def test_variable_role_default_value(metadata: Datadoc):
    assert all(
        v.variable_role == VariableRole.MEASURE.value for v in metadata.variables
    )


def test_is_personal_data_value(metadata: Datadoc):
    assert all(not v.is_personal_data for v in metadata.variables)


def test_save_file_path_metadata_field(
    existing_metadata_file: Path,
    metadata: Datadoc,
):
    metadata.write_metadata_document()
    with existing_metadata_file.open() as f:
        saved_file_path = json.load(f)["datadoc"]["dataset"]["file_path"]
    assert saved_file_path == str(metadata.dataset_path)


def test_save_file_path_dataset_and_no_metadata(
    metadata: Datadoc,
    tmp_path: pathlib.Path,
):
    metadata.write_metadata_document()
    with (tmp_path / TEST_EXISTING_METADATA_FILE_NAME).open() as f:
        saved_file_path = json.load(f)["datadoc"]["dataset"]["file_path"]
    assert saved_file_path == str(metadata.dataset_path)


@pytest.mark.parametrize(
    ("insert_string", "expected_from", "expected_until"),
    [
        ("_p2021", arrow.get("2021-01-01").date(), arrow.get("2021-12-31").date()),
        (
            "_p2022_p2023",
            arrow.get("2022-01-01").date(),
            arrow.get("2023-12-31").date(),
        ),
    ],
)
def test_period_metadata_fields_saved(
    subject_mapping_fake_statistical_structure: StatisticSubjectMapping,
    generate_periodic_file,
    expected_from,
    expected_until,
):
    metadata = Datadoc(
        str(generate_periodic_file),
        statistic_subject_mapping=subject_mapping_fake_statistical_structure,
    )
    assert metadata.dataset.contains_data_from == expected_from
    assert metadata.dataset.contains_data_until == expected_until


@pytest.mark.parametrize(
    ("dataset_path", "expected_type"),
    [
        (
            TEST_PROCESSED_DATA_POPULATION_DIRECTORY
            / "person_testdata_p2021-12-31_p2021-12-31_v1.parquet",
            DataSetStatus.INTERNAL.value,
        ),
        (
            TEST_PARQUET_FILEPATH,
            DataSetStatus.DRAFT.value,
        ),
        (
            "",
            None,
        ),
    ],
)
def test_dataset_status_default_value(
    subject_mapping_fake_statistical_structure: StatisticSubjectMapping,
    dataset_path: str,
    expected_type: DataSetStatus | None,
):
    datadoc_metadata = Datadoc(
        str(dataset_path),
        statistic_subject_mapping=subject_mapping_fake_statistical_structure,
    )
    assert datadoc_metadata.dataset.dataset_status == expected_type


@pytest.mark.parametrize(
    ("path_parts_to_insert", "expected_type"),
    [
        (
            "kildedata",
            Assessment.SENSITIVE.value,
        ),
        (
            "inndata",
            Assessment.PROTECTED.value,
        ),
        (
            "klargjorte_data",
            Assessment.PROTECTED.value,
        ),
        (
            "statistikk",
            Assessment.PROTECTED.value,
        ),
        (
            "utdata",
            Assessment.OPEN.value,
        ),
        (
            "",
            None,
        ),
    ],
)
def test_dataset_assessment_default_value(
    expected_type: Assessment | None,
    copy_dataset_to_path: ReadablePathLike,
    thread_pool_executor,
):
    datadoc_metadata = Datadoc(
        dataset_path=str(copy_dataset_to_path),
        statistic_subject_mapping=StatisticSubjectMapping(
            thread_pool_executor,
            source_url="",
        ),
    )
    assert datadoc_metadata.dataset.assessment == expected_type


@pytest.mark.parametrize(
    ("path_parts_to_insert", "expected_subject_code"),
    [
        (["aa_kortnvan_01", "klargjorte_data"], "aa01"),
        (["ab_kortnvan", "utdata"], "ab00"),
        (["aa_kortnvan_01", "no_dataset_state"], None),
        (["unknown_short_name", "klargjorte_data"], None),
    ],
)
def test_extract_subject_field_value_from_statistic_structure_xml(
    subject_mapping_fake_statistical_structure: StatisticSubjectMapping,
    copy_dataset_to_path: ReadablePathLike,
    expected_subject_code: str,
):
    subject_mapping_fake_statistical_structure.wait_for_external_result()
    metadata = Datadoc(
        str(copy_dataset_to_path),
        statistic_subject_mapping=subject_mapping_fake_statistical_structure,
    )
    assert metadata.dataset.subject_field == expected_subject_code


def test_generate_variables_id(
    metadata: Datadoc,
):
    assert all(isinstance(v.id, UUID) for v in metadata.variables)


@pytest.mark.parametrize(
    "existing_metadata_path",
    [TEST_EXISTING_METADATA_DIRECTORY / "invalid_id_field"],
)
def test_existing_metadata_variables_none_id(
    existing_metadata_file: Path,
    metadata: Datadoc,
):
    with existing_metadata_file.open() as f:
        pre_open_id: list = [v["id"] for v in json.load(f)["datadoc"]["variables"]]
    assert all(i is None for i in pre_open_id)

    assert all(isinstance(v.id, UUID) for v in metadata.variables)

    metadata.write_metadata_document()
    with existing_metadata_file.open() as f:
        post_write_id: list = [v["id"] for v in json.load(f)["datadoc"]["variables"]]

    assert post_write_id == [str(v.id) for v in metadata.variables]


@pytest.mark.parametrize(
    "existing_metadata_path",
    [TEST_EXISTING_METADATA_DIRECTORY / "valid_variable_id_field"],
)
def test_existing_metadata_variables_valid_id(
    existing_metadata_file: Path,
    metadata: Datadoc,
):
    with existing_metadata_file.open() as f:
        pre_open_id: list = [v["id"] for v in json.load(f)["datadoc"]["variables"]]

    assert all(isinstance(v.id, UUID) for v in metadata.variables)
    metadata_variable_ids = [str(v.id) for v in metadata.variables]
    assert metadata_variable_ids == pre_open_id

    metadata.write_metadata_document()
    with existing_metadata_file.open() as f:
        post_write_id: list = [v["id"] for v in json.load(f)["datadoc"]["variables"]]

    assert pre_open_id == post_write_id


@pytest.mark.parametrize(
    ("index", "expected_text"),
    [
        (0, "Norge"),
        (1, "Noreg"),
        (2, "Norway"),
    ],
)
def test_default_spatial_coverage_description(
    metadata: Datadoc,
    index: int,
    expected_text: str,
):
    ls = metadata.dataset.spatial_coverage_description
    assert ls.root[index].languageText == expected_text  # type: ignore[union-attr, index]


def test_open_extracted_and_existing_metadata(metadata_merged: Datadoc, tmp_path: Path):
    assert (
        metadata_merged.metadata_document
        == tmp_path
        / "ifpn/klargjorte_data/person_testdata_p2021-12-31_p2021-12-31_v1__DOC.json"
    )
    assert str(metadata_merged.dataset_path) is not None


def test_open_nonexistent_existing_metadata(existing_data_path: UPath):
    with pytest.raises(
        ValueError,
        match="Metadata document does not exist! Provided path:",
    ):
        Datadoc(
            str(existing_data_path),
            str(Datadoc.build_metadata_document_path(existing_data_path)),
        )


def test_merge_extracted_and_existing_dataset_metadata(metadata_merged: Datadoc):
    metadata_extracted = Datadoc(
        dataset_path=str(metadata_merged.dataset_path),
    )
    metadata_existing = Datadoc(
        metadata_document_path=str(TEST_EXISTING_METADATA_NAMING_STANDARD_FILEPATH),
    )

    # Should match extracted metadata from the dataset
    assert metadata_merged.dataset.short_name == metadata_extracted.dataset.short_name
    assert metadata_merged.dataset.assessment == metadata_extracted.dataset.assessment
    assert (
        metadata_merged.dataset.dataset_state
        == metadata_extracted.dataset.dataset_state
    )
    assert metadata_merged.dataset.version == metadata_extracted.dataset.version
    assert metadata_merged.dataset.file_path == metadata_extracted.dataset.file_path
    assert (
        metadata_merged.dataset.metadata_created_by
        == metadata_extracted.dataset.metadata_created_by
    )
    assert (
        metadata_merged.dataset.metadata_last_updated_by
        == metadata_extracted.dataset.metadata_last_updated_by
    )
    assert (
        metadata_merged.dataset.contains_data_from
        == metadata_extracted.dataset.contains_data_from
    )
    assert (
        metadata_merged.dataset.contains_data_until
        == metadata_extracted.dataset.contains_data_until
    )

    # Should match existing metadata
    for field in DATASET_FIELDS_FROM_EXISTING_METADATA:
        assert getattr(metadata_merged.dataset, field) == getattr(
            metadata_existing.dataset,
            field,
        ), f"{field} in merged metadata did not match existing metadata"

    # Special cases
    assert metadata_merged.dataset.id != metadata_existing.dataset.id
    assert metadata_merged.dataset.metadata_created_date is None
    assert metadata_merged.dataset.metadata_last_updated_date is None


def test_merge_variables(tmp_path):
    dataset = tmp_path / "fewer_variables_p2021-12-31_p2021-12-31_v1.parquet"
    existing_document = TEST_EXISTING_METADATA_NAMING_STANDARD_FILEPATH
    dataset.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(
        TEST_DATASETS_DIRECTORY / "fewer_variables_p2021-12-31_p2021-12-31_v1.parquet",
        dataset,
    )
    extracted = Datadoc(
        dataset_path=str(dataset),
    )
    existing = Datadoc(
        metadata_document_path=str(existing_document),
    )
    merged = Datadoc(
        dataset_path=str(dataset),
        metadata_document_path=str(existing_document),
        errors_as_warnings=True,
    )
    assert [v.short_name for v in merged.variables] == [
        v.short_name for v in extracted.variables
    ]
    assert all(v.id is not None for v in merged.variables)
    assert [v.id for v in merged.variables] != [v.id for v in existing.variables]
    assert all(
        v.contains_data_from == merged.dataset.contains_data_from
        for v in merged.variables
    )
    assert all(
        v.contains_data_until == merged.dataset.contains_data_until
        for v in merged.variables
    )


def test_merge_with_fewer_variables_in_existing_metadata(tmp_path):
    target = tmp_path / TEST_NAMING_STANDARD_COMPATIBLE_DATASET
    target.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(
        TEST_DATASETS_DIRECTORY / TEST_NAMING_STANDARD_COMPATIBLE_DATASET,
        target,
    )
    datadoc = Datadoc(
        str(target),
        str(
            TEST_EXISTING_METADATA_DIRECTORY
            / "fewer_variables_p2020-12-31_p2020-12-31_v1__DOC.json",
        ),
        errors_as_warnings=True,
    )
    assert [v.short_name for v in datadoc.variables] == [
        "fnr",
        "sivilstand",
        "bostedskommune",
        "inntekt",
        "bankinnskudd",
        "dato",
    ]



from pathlib import Path

RESOURCES_DIR = Path.cwd() / "tests" / "variable_definitions" / "resources"
OPENAPI_DIR = RESOURCES_DIR / "openapi"
VARIABLE_DEFINITION_EDITING_FILES_DIR = (
    RESOURCES_DIR / "variable_definition_editing_files"
)



"""Tests for variable definition files."""

import logging
from datetime import datetime
from datetime import timedelta
from pathlib import Path

import pytest
import pytz
from pytest_mock import MockerFixture
from ruamel.yaml import YAML

from dapla_metadata.variable_definitions._generated.vardef_client.models.variable_status import (
    VariableStatus,
)
from dapla_metadata.variable_definitions._utils.config import WORKSPACE_DIR
from dapla_metadata.variable_definitions._utils.constants import TEMPLATE_HEADER
from dapla_metadata.variable_definitions._utils.constants import (
    TEMPLATE_SECTION_HEADER_MACHINE_GENERATED,
)
from dapla_metadata.variable_definitions._utils.constants import (
    TEMPLATE_SECTION_HEADER_OWNER,
)
from dapla_metadata.variable_definitions._utils.constants import (
    TEMPLATE_SECTION_HEADER_STATUS,
)
from dapla_metadata.variable_definitions._utils.files import _get_workspace_dir
from dapla_metadata.variable_definitions._utils.files import configure_yaml
from dapla_metadata.variable_definitions._utils.template_files import (
    _find_latest_template_file,
)
from dapla_metadata.variable_definitions._utils.template_files import (
    create_template_yaml,
)
from dapla_metadata.variable_definitions.exceptions import VardefFileError
from dapla_metadata.variable_definitions.vardef import Vardef
from tests.utils.constants import VARDEF_EXAMPLE_DEFINITION_ID
from tests.utils.constants import VARDEF_EXAMPLE_SHORT_NAME
from tests.variable_definitions.conftest import VARIABLE_DEFINITION_DICT

yaml = YAML()
configure_yaml(yaml)


def test_yaml_file_creation(work_folder_defaults: Path):
    """Test if the function generates a YAML file."""
    file_path = work_folder_defaults
    assert file_path.exists(), "YAML file was not created"


def test_yaml_content_default_values(work_folder_defaults: Path):
    """Check if the generated YAML file with default values contains the expected data."""
    with work_folder_defaults.open(encoding="utf-8") as f:
        parsed_yaml = yaml.load(f)

    assert parsed_yaml["variable_status"] == "DRAFT"
    assert parsed_yaml["owner"]["team"] == "default team"
    assert parsed_yaml["name"]["nn"] is None


def test_yaml_content_saved_values(work_folder_variable_definition: Path) -> None:
    """Check if the generated YAML file with saved values contains the expected data."""
    with work_folder_variable_definition.open(encoding="utf-8") as f:
        parsed_yaml = yaml.load(f)

    assert parsed_yaml["variable_status"] == "PUBLISHED_EXTERNAL"
    assert parsed_yaml["last_updated_by"] == "ano@ssb.no"


def test_yaml_content_from_model_variable_definition(
    work_folder_variable_definition: Path,
) -> None:
    """Check if the generated YAML file with saved values contains the expected data."""
    with work_folder_variable_definition.open(encoding="utf-8") as f:
        parsed_yaml = yaml.load(f)

    assert parsed_yaml["variable_status"] == "PUBLISHED_EXTERNAL"
    assert parsed_yaml["last_updated_by"] == "ano@ssb.no"


def test_yaml_comments(work_folder_defaults: Path):
    """Ensure the YAML file includes the expected header comments."""
    with work_folder_defaults.open(encoding="utf-8") as f:
        content = f.read()

    assert TEMPLATE_HEADER in content
    assert TEMPLATE_SECTION_HEADER_STATUS.strip() in content.strip()
    assert TEMPLATE_SECTION_HEADER_OWNER.strip() in content.strip()
    assert TEMPLATE_SECTION_HEADER_MACHINE_GENERATED.strip() in content.strip()


def test_file_name(work_folder_defaults: Path):
    """Check filename."""
    assert "variable_definition_template_" in str(work_folder_defaults)


def test_workspace_fixture(set_temp_workspace: Path, tmp_path: Path):
    workspace = set_temp_workspace
    assert workspace.exists()
    assert workspace == tmp_path / "work"


@pytest.mark.usefixtures("set_temp_workspace")
def test_get_workspace_dir_without_env_var():
    workspace_dir = _get_workspace_dir()
    assert workspace_dir.exists()
    assert workspace_dir.is_dir()


def test_generate_yaml_from_dict() -> None:
    """Check if the generated YAML file with saved values contains the expected data."""
    with pytest.raises(AttributeError) as exc_info:
        create_template_yaml(
            VARIABLE_DEFINITION_DICT,  # type: ignore[arg-type]
            custom_directory=None,
        )
    assert str(exc_info.value) == "'dict' object has no attribute 'model_dump'"


def test_find_latest_template_file(mocker: MockerFixture, tmp_path: Path):
    mock_time = mocker.patch(
        "dapla_metadata.variable_definitions._utils.variable_definition_files._get_current_time",
    )

    def fast_forward(seconds: int) -> str:
        return str(
            datetime.now(pytz.timezone("Europe/Oslo")) + timedelta(seconds=seconds),
        )

    # Create three template files and one other
    # We need to fast-forward time to avoid all the files being named identically
    mock_time.return_value = fast_forward(0)
    create_template_yaml(custom_directory=tmp_path)
    mock_time.return_value = fast_forward(30)
    create_template_yaml(custom_directory=tmp_path)
    mock_time.return_value = fast_forward(60)
    third = create_template_yaml(custom_directory=tmp_path)
    (tmp_path / "random_file.yaml").touch()

    assert _find_latest_template_file(directory=tmp_path) == third


def test_find_latest_template_file_none_exist(tmp_path: Path):
    assert _find_latest_template_file(directory=tmp_path) is None


def test_logging_workspace_dir(caplog):
    """Test logging intended for debugging."""
    caplog.set_level(logging.DEBUG)
    _get_workspace_dir()
    assert "WORKSPACE_DIR' value:" in caplog.text


def test_workspace_dir_not_set(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.delenv(WORKSPACE_DIR, raising=False)
    with pytest.raises(
        VardefFileError,
        match="VardefFileError: WORKSPACE_DIR is not set",
    ):
        _get_workspace_dir()


def test_workspace_dir_doesnt_exist(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv(WORKSPACE_DIR, "funnypath/haha")
    with pytest.raises(FileNotFoundError):
        _get_workspace_dir()


def test_workspace_is_not_dir(monkeypatch: pytest.MonkeyPatch, set_workspace_not_dir):
    monkeypatch.setenv(WORKSPACE_DIR, str(set_workspace_not_dir))
    with pytest.raises(NotADirectoryError):
        _get_workspace_dir()


def test_write_existing_variable_to_file():
    file_name = (
        Vardef.get_variable_definition_by_id(
            variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
        )
        .to_file()
        .get_file_path()
    )
    with file_name.open(encoding="utf-8") as f:
        parsed_yaml = yaml.load(f)

    assert parsed_yaml["variable_status"] == VariableStatus.DRAFT
    assert parsed_yaml["short_name"] == VARDEF_EXAMPLE_SHORT_NAME


def test_write_existing_variable_to_file_by_short_name():
    file_name = (
        Vardef.get_variable_definition_by_shortname(
            short_name=VARDEF_EXAMPLE_SHORT_NAME,
        )
        .to_file()
        .get_file_path()
    )
    with file_name.open(encoding="utf-8") as f:
        parsed_yaml = yaml.load(f)

    assert parsed_yaml["variable_status"] == VariableStatus.DRAFT
    assert parsed_yaml["short_name"] == VARDEF_EXAMPLE_SHORT_NAME


def test_shortname_and_id_in_filename():
    file_name = (
        Vardef.get_variable_definition_by_id(
            variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
        )
        .to_file()
        .get_file_path()
    )

    assert (
        f"variable_definition_{VARDEF_EXAMPLE_SHORT_NAME}_{VARDEF_EXAMPLE_DEFINITION_ID}_"
        in str(file_name)
    )



"""Shared fixtures and configuration."""

from __future__ import annotations

import copy
import functools
import os
import shutil
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from datetime import timezone
from typing import TYPE_CHECKING

import pandas as pd
import pytest
from bs4 import BeautifulSoup
from bs4 import ResultSet
from upath import UPath

from dapla_metadata.dapla.user_info import TestUserInfo
from dapla_metadata.datasets import Datadoc
from dapla_metadata.datasets.code_list import CodeList
from dapla_metadata.datasets.statistic_subject_mapping import StatisticSubjectMapping
from dapla_metadata.datasets.utility.enums import SupportedLanguages
from tests.datasets.constants import CODE_LIST_DIR
from tests.datasets.constants import DATADOC_METADATA_MODULE
from tests.datasets.constants import TEST_DATASETS_DIRECTORY
from tests.datasets.constants import TEST_EXISTING_METADATA_DIRECTORY
from tests.datasets.constants import TEST_EXISTING_METADATA_FILE_NAME
from tests.datasets.constants import TEST_EXISTING_METADATA_NAMING_STANDARD_FILEPATH
from tests.datasets.constants import TEST_NAMING_STANDARD_COMPATIBLE_DATASET
from tests.datasets.constants import TEST_PARQUET_FILE_NAME
from tests.datasets.constants import TEST_PARQUET_FILEPATH
from tests.datasets.constants import TEST_RESOURCES_DIRECTORY
from tests.datasets.test_statistic_subject_mapping import (
    STATISTICAL_SUBJECT_STRUCTURE_DIR,
)

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


@pytest.fixture(autouse=True)
def _clear_environment(mocker: MockerFixture) -> None:
    """Ensure that the environment is cleared."""
    mocker.patch.dict(os.environ, clear=True)


@pytest.fixture(scope="session", autouse=True)
def faker_session_locale():
    return ["no_NO"]


@pytest.fixture
def dummy_timestamp() -> datetime:
    return datetime(2022, 1, 1, tzinfo=timezone.utc)


@pytest.fixture
def _mock_timestamp(mocker: MockerFixture, dummy_timestamp: datetime) -> None:
    mocker.patch(
        DATADOC_METADATA_MODULE + ".core.get_timestamp_now",
        return_value=dummy_timestamp,
    )


@pytest.fixture
def _mock_user_info(mocker: MockerFixture) -> None:
    mocker.patch(
        "dapla_metadata.dapla.user_info.get_user_info_for_current_platform",
        return_value=TestUserInfo(),
    )


@pytest.fixture
def metadata(
    _mock_timestamp: None,
    _mock_user_info: None,
    subject_mapping_fake_statistical_structure: StatisticSubjectMapping,
    tmp_path: UPath,
) -> Datadoc:
    shutil.copy(str(TEST_PARQUET_FILEPATH), str(tmp_path / TEST_PARQUET_FILE_NAME))
    return Datadoc(
        str(tmp_path / TEST_PARQUET_FILE_NAME),
        statistic_subject_mapping=subject_mapping_fake_statistical_structure,
    )


@pytest.fixture
def metadata_merged(
    _mock_timestamp: None,
    _mock_user_info: None,
    subject_mapping_fake_statistical_structure: StatisticSubjectMapping,
    tmp_path: UPath,
) -> Datadoc:
    target = tmp_path / TEST_NAMING_STANDARD_COMPATIBLE_DATASET
    target.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(
        str(TEST_DATASETS_DIRECTORY / TEST_NAMING_STANDARD_COMPATIBLE_DATASET),
        str(target),
    )
    return Datadoc(
        str(target),
        str(TEST_EXISTING_METADATA_NAMING_STANDARD_FILEPATH),
        statistic_subject_mapping=subject_mapping_fake_statistical_structure,
        errors_as_warnings=True,
    )


@pytest.fixture
def existing_metadata_path() -> UPath:
    return TEST_EXISTING_METADATA_DIRECTORY


@pytest.fixture
def existing_metadata_file(tmp_path: UPath, existing_metadata_path: UPath) -> UPath:
    # Setup by copying the file into the relevant directory
    shutil.copy(
        str(existing_metadata_path / TEST_EXISTING_METADATA_FILE_NAME),
        str(tmp_path / TEST_EXISTING_METADATA_FILE_NAME),
    )
    return tmp_path / TEST_EXISTING_METADATA_FILE_NAME


@pytest.fixture
def english_name() -> str:
    return "English Name"


@pytest.fixture
def bokmål_name() -> str:  # noqa: PLC2401 the listed problems do not apply in this case
    return "Bokmål navn"


@pytest.fixture
def nynorsk_name() -> str:
    return "Nynorsk namn"


@pytest.fixture
def existing_data_path() -> UPath:
    return TEST_PARQUET_FILEPATH


@pytest.fixture
def full_dataset_state_path(
    path_parts_to_insert: str | list[str],
) -> UPath:
    """Create a longer path structure from just one section.

    Examples:
    >>> full_dataset_state_path('inndata')
    'tests/dataset/inndata/resources/person_data_v1.parquet'
    >>> full_dataset_state_path(['klargjorte_data', 'arbmark'])
    'tests/dataset/klargjorte_data/arbmark/resources/person_data_v1.parquet'
    """
    split_path = list(UPath(TEST_PARQUET_FILEPATH).parts)
    new_path = copy.copy(split_path)

    if isinstance(path_parts_to_insert, str):
        parts = [path_parts_to_insert]
    else:
        parts = path_parts_to_insert
    for p in parts:
        new_path.insert(-2, p)
    return UPath().joinpath(*new_path)


@pytest.fixture
def subject_xml_file_path() -> UPath:
    return (
        TEST_RESOURCES_DIRECTORY
        / STATISTICAL_SUBJECT_STRUCTURE_DIR
        / "extract_secondary_subject.xml"
    )


@pytest.fixture
def thread_pool_executor() -> ThreadPoolExecutor:
    return ThreadPoolExecutor(max_workers=12)


@pytest.fixture
def subject_mapping_fake_statistical_structure(
    _mock_fetch_statistical_structure,
    thread_pool_executor,
) -> StatisticSubjectMapping:
    return StatisticSubjectMapping(thread_pool_executor, "placeholder")


@pytest.fixture
def _mock_fetch_statistical_structure(
    mocker,
    subject_xml_file_path: UPath,
) -> None:
    def fake_statistical_structure(self=None) -> ResultSet:  # noqa: ARG001
        with subject_xml_file_path.open() as f:
            return BeautifulSoup(f.read(), features="xml").find_all("hovedemne")

    mocker.patch(
        DATADOC_METADATA_MODULE
        + ".statistic_subject_mapping.StatisticSubjectMapping._fetch_data_from_external_source",
        functools.partial(fake_statistical_structure),
    )


@pytest.fixture
def subject_mapping_http_exception(
    requests_mock,
    exception_to_raise,
    thread_pool_executor,
) -> StatisticSubjectMapping:
    requests_mock.get(
        "http://test.some.url.com",
        exc=exception_to_raise,
    )
    return StatisticSubjectMapping(thread_pool_executor, "http://test.some.url.com")


@pytest.fixture
def code_list_csv_filepath_nb() -> UPath:
    return TEST_RESOURCES_DIRECTORY / CODE_LIST_DIR / "code_list_nb.csv"


@pytest.fixture
def code_list_csv_filepath_nn() -> UPath:
    return TEST_RESOURCES_DIRECTORY / CODE_LIST_DIR / "code_list_nn.csv"


@pytest.fixture
def code_list_csv_filepath_en() -> UPath:
    return TEST_RESOURCES_DIRECTORY / CODE_LIST_DIR / "code_list_en.csv"


@pytest.fixture
def _mock_fetch_dataframe(
    mocker,
    code_list_csv_filepath_nb: UPath,
    code_list_csv_filepath_nn: UPath,
    code_list_csv_filepath_en: UPath,
) -> None:
    def fake_code_list(self=None) -> dict[SupportedLanguages, pd.DataFrame]:  # noqa: ARG001
        return {
            SupportedLanguages.NORSK_BOKMÅL: pd.read_csv(
                str(code_list_csv_filepath_nb),
                converters={"code": str},
            ),
            SupportedLanguages.NORSK_NYNORSK: pd.read_csv(
                str(code_list_csv_filepath_nn),
                converters={"code": str},
            ),
            SupportedLanguages.ENGLISH: pd.read_csv(
                str(code_list_csv_filepath_en),
                converters={"code": str},
            ),
        }

    mocker.patch(
        DATADOC_METADATA_MODULE
        + ".code_list.CodeList._fetch_data_from_external_source",
        functools.partial(fake_code_list),
    )


@pytest.fixture
def code_list_fake_structure(_mock_fetch_dataframe, thread_pool_executor) -> CodeList:
    return CodeList(thread_pool_executor, 100)


@pytest.fixture
def copy_dataset_to_path(
    tmp_path: UPath,
    full_dataset_state_path: UPath,
) -> UPath:
    temporary_dataset = tmp_path / full_dataset_state_path
    temporary_dataset.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(str(TEST_PARQUET_FILEPATH), str(temporary_dataset))
    return temporary_dataset



"""Repository for constant values for Datadoc tests."""

from datadoc_model.all_optional.model import DataType
from upath import UPath

TEST_BUCKET_PARQUET_FILEPATH = "gs://ssb-staging-dapla-felles-data-delt/datadoc/klargjorte_data/person_data_v1.parquet"

TEST_BUCKET_PARQUET_FILEPATH_WITH_SHORTNAME = "gs://ssb-staging-dapla-felles-data-delt/befolkning/klargjorte_data/person_data_v1.parquet"

TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH = "gs://ssb-my-team-data-produkt-prod/ifpn/klargjorte_data/person_testdata_p2021-12-31_p2021-12-31_v1.parquet"

TEST_RESOURCES_DIRECTORY = UPath("tests/datasets/resources")

TEST_DATASETS_DIRECTORY = TEST_RESOURCES_DIRECTORY / "datasets"

TEST_PARQUET_FILE_NAME = "person_data_v1.parquet"

TEST_NAMING_STANDARD_COMPATIBLE_DATASET = (
    "ifpn/klargjorte_data/person_testdata_p2021-12-31_p2021-12-31_v1.parquet"
)

TEST_PARQUET_FILEPATH = TEST_DATASETS_DIRECTORY / TEST_PARQUET_FILE_NAME

TEST_SAS7BDAT_FILEPATH = TEST_DATASETS_DIRECTORY / "sasdata.sas7bdat"

TEST_PARQUET_GZIP_FILEPATH = TEST_DATASETS_DIRECTORY / "person_data_v1.parquet.gzip"

TEST_EXISTING_METADATA_DIRECTORY = TEST_RESOURCES_DIRECTORY / "existing_metadata_file"

TEST_EXISTING_METADATA_FILE_NAME = "person_data_v1__DOC.json"

TEST_EXISTING_METADATA_FILEPATH = (
    TEST_EXISTING_METADATA_DIRECTORY / TEST_EXISTING_METADATA_FILE_NAME
)

TEST_EXISTING_METADATA_NAMING_STANDARD_FILEPATH = (
    TEST_EXISTING_METADATA_DIRECTORY
    / "person_testdata_p2020-12-31_p2020-12-31_v1__DOC.json"
)

TEST_EXISTING_METADATA_WITH_VALID_ID_DIRECTORY = (
    TEST_EXISTING_METADATA_DIRECTORY / "valid_id_field"
)

TEST_COMPATIBILITY_DIRECTORY = TEST_EXISTING_METADATA_DIRECTORY / "compatibility"

TEST_PSEUDO_DIRECTORY = TEST_EXISTING_METADATA_DIRECTORY / "pseudo"

TEST_PROCESSED_DATA_POPULATION_DIRECTORY = TEST_RESOURCES_DIRECTORY / "klargjorte_data"


CODE_LIST_DIR = "code_list"

DATADOC_METADATA_MODULE = "dapla_metadata.datasets"

DATADOC_METADATA_MODULE_CORE = "dapla_metadata.datasets.core"

DATADOC_METADATA_MODULE_UTILS = "dapla_metadata.datasets.utility.utils"

VARIABLE_SHORT_NAMES = [
    "fnr",
    "sivilstand",
    "bostedskommune",
    "inntekt",
    "bankinnskudd",
    "dato",
]


VARIABLE_DATA_TYPES = [
    DataType.STRING,
    DataType.STRING,
    DataType.STRING,
    DataType.INTEGER,
    DataType.INTEGER,
    DataType.DATETIME,
]



from __future__ import annotations

import contextlib
import logging
from typing import Protocol

import jwt

from dapla_metadata._shared import config
from dapla_metadata._shared.enums import DaplaRegion

logger = logging.getLogger(__name__)


class UserInfo(Protocol):
    """Information about the current user.

    Implementations may be provided for different platforms or testing.
    """

    @property
    def short_email(self) -> str | None:
        """Get the short email address."""
        ...

    @property
    def current_group(self) -> str:
        """Get the group which the user is currently representing."""
        ...

    @property
    def current_team(self) -> str:
        """Get the team which the user is currently representing."""
        ...


class UnknownUserInfo:
    """Fallback when no implementation is found."""

    @property
    def short_email(self) -> str | None:
        """Unknown email address."""
        return None

    @property
    def current_group(self) -> str:
        """Get the group which the user is currently representing."""
        return ""

    @property
    def current_team(self) -> str:
        """Get the team which the user is currently representing."""
        return ""


class TestUserInfo:
    """Information about the current user for local development and testing."""

    PLACEHOLDER_EMAIL_ADDRESS = "default_user@ssb.no"
    PLACEHOLDER_GROUP = "default-team-developers"
    PLACEHOLDER_TEAM = "default-team"

    @property
    def short_email(self) -> str | None:
        """Get the short email address."""
        return TestUserInfo.PLACEHOLDER_EMAIL_ADDRESS

    @property
    def current_group(self) -> str | None:
        """Get the group which the user is currently representing."""
        return TestUserInfo.PLACEHOLDER_GROUP

    @property
    def current_team(self) -> str | None:
        """Get the team which the user is currently representing."""
        return TestUserInfo.PLACEHOLDER_TEAM


class DaplaLabUserInfo:
    """Information about the current user when running on Dapla Lab."""

    @property
    def short_email(self) -> str | None:
        """Get the short email address."""
        encoded_jwt = config.get_oidc_token()
        if encoded_jwt:
            # The JWT has been verified by the platform prior to injection, no need to verify.
            decoded_jwt = jwt.decode(encoded_jwt, options={"verify_signature": False})
            with contextlib.suppress(KeyError):
                # If email can't be found in the JWT, fall through and return None
                return decoded_jwt["email"]

        logger.warning(
            "Could not access JWT from environment. Could not get short email address.",
        )
        return None

    @property
    def current_group(self) -> str:
        """Get the group which the user is currently representing."""
        if group := config.get_dapla_group_context():
            return group
        msg = "DAPLA_GROUP_CONTEXT environment variable not found"
        raise OSError(msg)

    @property
    def current_team(self) -> str:
        """Get the team which the user is currently representing."""
        return parse_team_name(self.current_group)


def get_user_info_for_current_platform() -> UserInfo:
    """Return the correct implementation of UserInfo for the current platform."""
    if config.get_dapla_region() == DaplaRegion.DAPLA_LAB:
        return DaplaLabUserInfo()
    logger.warning(
        "Was not possible to retrieve user information! Some fields may not be set.",
    )
    return UnknownUserInfo()


def parse_team_name(group: str) -> str:
    """Parses the group to get the current team.

    >>> parse_team_name("dapla-metadata-developers")
    'dapla-metadata'

    >>> parse_team_name("dapla-metadata-data-admins")
    'dapla-metadata'

    >>> parse_team_name("dapla-metadata")
    'dapla'

    >>> parse_team_name("dapla-metadata-not-real-name")
    'dapla-metadata-not-real'
    """
    parts = group.split("-")
    return "-".join(parts[:-2] if group.endswith("data-admins") else parts[:-1])



"""Tests for the compatibility package."""

import json
from pathlib import Path

import pytest
from upath import UPath

from dapla_metadata.datasets.compatibility import is_metadata_in_container_structure
from dapla_metadata.datasets.compatibility import upgrade_metadata
from dapla_metadata.datasets.compatibility._handlers import handle_version_2_2_0
from dapla_metadata.datasets.compatibility._handlers import handle_version_3_3_0
from dapla_metadata.datasets.compatibility._handlers import handle_version_4_0_0
from dapla_metadata.datasets.compatibility._handlers import handle_version_5_0_1
from dapla_metadata.datasets.compatibility._handlers import handle_version_6_0_0
from dapla_metadata.datasets.compatibility._utils import DATADOC_KEY
from dapla_metadata.datasets.compatibility._utils import DATASET_KEY
from dapla_metadata.datasets.compatibility._utils import DOCUMENT_VERSION_KEY
from dapla_metadata.datasets.compatibility._utils import PSEUDONYMIZATION_KEY
from dapla_metadata.datasets.compatibility._utils import VARIABLES_KEY
from dapla_metadata.datasets.compatibility._utils import UnknownModelVersionError
from dapla_metadata.datasets.compatibility._utils import add_container
from dapla_metadata.datasets.compatibility._utils import convert_is_personal_data
from dapla_metadata.datasets.compatibility._utils import copy_pseudonymization_metadata
from dapla_metadata.datasets.core import Datadoc
from tests.datasets.constants import TEST_COMPATIBILITY_DIRECTORY
from tests.datasets.constants import TEST_EXISTING_METADATA_FILE_NAME
from tests.datasets.constants import TEST_PSEUDO_DIRECTORY

BACKWARDS_COMPATIBLE_VERSION_DIRECTORIES = [
    d for d in TEST_COMPATIBILITY_DIRECTORY.iterdir() if d.is_dir()
]

BACKWARDS_COMPATIBLE_VERSION_NAMES = [
    d.stem for d in BACKWARDS_COMPATIBLE_VERSION_DIRECTORIES
]


def test_existing_metadata_current_model_version():
    current_model_version = "6.1.0"
    fresh_metadata = {DATADOC_KEY: {DOCUMENT_VERSION_KEY: current_model_version}}
    upgraded_metadata = upgrade_metadata(fresh_metadata)
    assert upgraded_metadata == fresh_metadata


def test_handle_version_2_2_0() -> None:
    pydir: UPath = UPath(__file__).resolve().parent
    rootdir: UPath = pydir.parent.parent
    existing_metadata_file: UPath = (
        rootdir
        / TEST_COMPATIBILITY_DIRECTORY
        / "v2_2_0"
        / TEST_EXISTING_METADATA_FILE_NAME
    )
    with existing_metadata_file.open(mode="r", encoding="utf-8") as file:
        fresh_metadata = json.load(file)
    upgraded_metadata = handle_version_2_2_0(fresh_metadata)
    assert "custom_type" in upgraded_metadata[DATADOC_KEY][DATASET_KEY]
    assert "custom_type" in upgraded_metadata[DATADOC_KEY][VARIABLES_KEY][0]
    assert "special_value" in upgraded_metadata[DATADOC_KEY][VARIABLES_KEY][0]


def test_handle_version_3_3_0() -> None:
    pydir: UPath = UPath(__file__).resolve().parent
    rootdir: UPath = pydir.parent.parent
    existing_metadata_file: UPath = (
        rootdir
        / TEST_COMPATIBILITY_DIRECTORY
        / "v3_3_0"
        / TEST_EXISTING_METADATA_FILE_NAME
    )
    with existing_metadata_file.open(mode="r", encoding="utf-8") as file:
        fresh_metadata = json.load(file)
    upgraded_metadata = handle_version_3_3_0(fresh_metadata)
    assert (
        "direct_person_identifying"
        not in upgraded_metadata[DATADOC_KEY][VARIABLES_KEY][0]
    )


def test_handle_version_4_0_0() -> None:
    pydir: UPath = UPath(__file__).resolve().parent
    rootdir: UPath = pydir.parent.parent
    existing_metadata_file: UPath = (
        rootdir
        / TEST_PSEUDO_DIRECTORY
        / "dataset_and_pseudo"
        / "v4_0_0_person_data_v1__DOC.json"
    )
    with existing_metadata_file.open(mode="r", encoding="utf-8") as file:
        fresh_metadata = json.load(file)
    upgraded_metadata = handle_version_4_0_0(fresh_metadata)
    pseudo_field = upgraded_metadata[DATADOC_KEY][VARIABLES_KEY][0][
        PSEUDONYMIZATION_KEY
    ]
    assert pseudo_field["encryption_algorithm"] == "TINK-DAEAD"
    assert pseudo_field["encryption_key_reference"] == "ssb-common-key-1"
    assert pseudo_field["pseudonymization_time"] == "2010-09-05"
    assert pseudo_field["encryption_algorithm_parameters"] == [
        {"keyId": "ssb-common-key-1"}
    ]
    assert (
        upgraded_metadata[DATADOC_KEY][VARIABLES_KEY][1][PSEUDONYMIZATION_KEY] is None
    )


def test_handle_version_4_0_0_without_pseudo() -> None:
    pydir: UPath = UPath(__file__).resolve().parent
    rootdir: UPath = pydir.parent.parent
    existing_metadata_file: UPath = (
        rootdir
        / TEST_PSEUDO_DIRECTORY
        / "dataset_and_pseudo"
        / "no_pseudo_v4_0_0_person_data_v1__DOC.json"
    )
    with existing_metadata_file.open(mode="r", encoding="utf-8") as file:
        fresh_metadata = json.load(file)
    upgraded_metadata = handle_version_4_0_0(fresh_metadata)
    assert PSEUDONYMIZATION_KEY not in upgraded_metadata
    assert all(
        PSEUDONYMIZATION_KEY not in v
        for v in upgraded_metadata[DATADOC_KEY][VARIABLES_KEY]
    )


def test_handle_version_5_0_1() -> None:
    pydir: UPath = UPath(__file__).resolve().parent
    rootdir: UPath = pydir.parent.parent
    existing_metadata_file: UPath = (
        rootdir
        / TEST_COMPATIBILITY_DIRECTORY
        / "v5_0_1"
        / TEST_EXISTING_METADATA_FILE_NAME
    )
    with existing_metadata_file.open(mode="r", encoding="utf-8") as file:
        fresh_metadata = json.load(file)
    upgraded_metadata = handle_version_5_0_1(fresh_metadata)
    removed_dataset_fields = [
        "contains_personal_data",
        "unit_type",
        "data_source",
        "temporality_type",
    ]
    for f in removed_dataset_fields:
        assert f not in upgraded_metadata[DATADOC_KEY][DATASET_KEY]
    added_variables_fields = [
        "is_personal_data",
        "unit_type",
        "data_source",
        "temporality_type",
    ]
    for f in added_variables_fields:
        assert all(f in v for v in upgraded_metadata[DATADOC_KEY][VARIABLES_KEY])


def test_handle_version_6_0_0() -> None:
    existing_metadata_file = (
        Path(__file__).resolve().parents[2]
        / TEST_COMPATIBILITY_DIRECTORY
        / "v6_0_0"
        / TEST_EXISTING_METADATA_FILE_NAME
    )
    fresh_metadata = json.loads(existing_metadata_file.read_text(encoding="utf-8"))

    upgraded = handle_version_6_0_0(fresh_metadata)

    dataset = upgraded[DATADOC_KEY][DATASET_KEY]

    assert all(f not in dataset for f in ["use_restriction", "use_restriction_date"])

    assert "use_restrictions" in dataset

    first_use_restriction = dataset["use_restrictions"][0]
    assert first_use_restriction["use_restriction_type"] == "DELETION_ANONYMIZATION"
    assert first_use_restriction["use_restriction_date"] == "2022-09-05"


def test_existing_metadata_unknown_model_version():
    fresh_metadata = {DATADOC_KEY: {DOCUMENT_VERSION_KEY: "0.27.65"}}
    with pytest.raises(UnknownModelVersionError):
        upgrade_metadata(fresh_metadata)


@pytest.mark.parametrize(
    "existing_metadata_path",
    BACKWARDS_COMPATIBLE_VERSION_DIRECTORIES,
    ids=BACKWARDS_COMPATIBLE_VERSION_NAMES,
)
def test_backwards_compatibility(
    existing_metadata_file: Path,
    metadata: Datadoc,
):
    with existing_metadata_file.open() as f:
        file_metadata = json.loads(f.read())

    if is_metadata_in_container_structure(file_metadata):
        file_metadata = file_metadata[DATADOC_KEY]

    # Just test a single value to make sure we have a working model
    assert metadata.dataset.short_name == file_metadata[DATASET_KEY]["short_name"]  # type: ignore [union-attr, index]


def test_add_container():
    doc = {
        "percentage_complete": 98,
        DOCUMENT_VERSION_KEY: "2.1.0",
        DATASET_KEY: {"short_name": "person_data_v1", "assessment": "SENSITIVE"},
    }
    doc_with_container = add_container(doc)
    assert doc_with_container[DOCUMENT_VERSION_KEY] == "0.0.1"
    assert doc_with_container[DATADOC_KEY][DOCUMENT_VERSION_KEY] == "2.1.0"
    assert PSEUDONYMIZATION_KEY in doc_with_container


@pytest.mark.parametrize(
    ("input_value", "expected_result"),
    [
        ("NON_PSEUDONYMISED_ENCRYPTED_PERSONAL_DATA", True),
        ("PSEUDONYMISED_ENCRYPTED_PERSONAL_DATA", True),
        ("NOT_PERSONAL_DATA", False),
        (None, None),
        (False, False),
        (True, True),
    ],
)
def test_convert_is_personal_data(input_value, expected_result):
    supplied_metadata = {
        DATADOC_KEY: {VARIABLES_KEY: [{"is_personal_data": input_value}]}
    }

    convert_is_personal_data(supplied_metadata)

    assert (
        supplied_metadata[DATADOC_KEY][VARIABLES_KEY][0]["is_personal_data"]
        == expected_result
    )


def test_copy_pseudonymization_metadata_shortname_mismatch():
    supplied_metadata = {
        DATADOC_KEY: {VARIABLES_KEY: [{"short_name": "pers_id"}]},
        PSEUDONYMIZATION_KEY: {
            DOCUMENT_VERSION_KEY: "0.1.0",
            "pseudo_dataset": None,
            "pseudo_variables": [
                {
                    "short_name": "fnr",
                    "data_element_path": "fnr",
                    "data_element_pattern": "**/fnr",
                    "stable_identifier_type": None,
                    "stable_identifier_version": None,
                    "encryption_algorithm": "TINK-DAEAD",
                    "encryption_key_reference": "ssb-common-key-1",
                    "encryption_algorithm_parameters": [{"keyId": "ssb-common-key-1"}],
                    "source_variable": None,
                    "source_variable_datatype": None,
                }
            ],
        },
    }

    copy_pseudonymization_metadata(supplied_metadata)

    assert len(supplied_metadata[DATADOC_KEY][VARIABLES_KEY]) == 1
    assert (
        supplied_metadata[DATADOC_KEY][VARIABLES_KEY][0][PSEUDONYMIZATION_KEY] is None
    )



"""Tests for Vardef client exception handling."""

from http import HTTPStatus

from dapla_metadata.variable_definitions.exceptions import STATUS_EXPLANATIONS
from dapla_metadata.variable_definitions.exceptions import VardefClientError
from tests.utils.constants import CONSTRAINT_VIOLATION_BODY
from tests.utils.constants import CONSTRAINT_VIOLATION_BODY_MISSING_FIELD
from tests.utils.constants import CONSTRAINT_VIOLATION_BODY_MISSING_MESSAGES
from tests.utils.constants import CONSTRAINT_VIOLATION_BODY_MISSING_VIOLATIONS


def test_valid_response_body():
    response_body = '{"status": 400, "detail": "Bad Request"}'
    exc = VardefClientError(response_body)
    assert exc.status == HTTPStatus.BAD_REQUEST
    assert exc.detail == "\nDetail: Bad Request"
    assert str(exc) == STATUS_EXPLANATIONS[HTTPStatus.BAD_REQUEST] + exc.detail


def test_response_empty_status():
    response_body = '{"status": , "detail": "Bad Request"}'
    exc = VardefClientError(response_body)
    assert exc.status is None
    assert str(exc) == "Could not decode error response from API"


def test_no_status():
    response_body = '{"detail": "Bad Request"}'
    exc = VardefClientError(response_body)
    assert exc.status is None
    assert exc.detail == "\nDetail: Bad Request"


def test_no_explanation_for_status():
    assert (
        str(
            VardefClientError(
                f'{{"status": {HTTPStatus.IM_A_TEAPOT}, "detail": "{HTTPStatus.IM_A_TEAPOT.phrase}"}}',
            ),
        )
        == f"Status {HTTPStatus.IM_A_TEAPOT}:\nDetail: {HTTPStatus.IM_A_TEAPOT.phrase}"
    )


def test_invalid_json():
    response_body = "Not a JSON string"
    exc = VardefClientError(response_body)
    assert exc.status is None
    assert exc.detail == "Could not decode error response from API"
    assert str(exc) == "Could not decode error response from API"


def test_missing_keys():
    response_body = '{"status": 404}'
    exc = VardefClientError(response_body)
    assert exc.status == HTTPStatus.NOT_FOUND
    assert exc.detail is None
    assert str(exc) == STATUS_EXPLANATIONS[HTTPStatus.NOT_FOUND]


def test_constraint_violation():
    response_body = CONSTRAINT_VIOLATION_BODY
    exc = VardefClientError(response_body)
    assert exc.status == HTTPStatus.BAD_REQUEST
    assert exc.detail == (
        "\nDetail: "
        "\nupdateVariableDefinitionById.updateDraft.owner.team: Invalid Dapla team"
        "\nupdateVariableDefinitionById.updateDraft.owner.team: must not be empty"
    )


def test_constraint_violation_missing_messages():
    response_body = CONSTRAINT_VIOLATION_BODY_MISSING_MESSAGES
    exc = VardefClientError(response_body)
    assert exc.status == HTTPStatus.BAD_REQUEST
    assert exc.detail == (
        "\nDetail: "
        "\nupdateVariableDefinitionById.updateDraft.owner.team: No message provided"
        "\nupdateVariableDefinitionById.updateDraft.owner.team: No message provided"
    )


def test_constraint_violation_empty_violations():
    response_body = CONSTRAINT_VIOLATION_BODY_MISSING_VIOLATIONS
    exc = VardefClientError(response_body)
    assert exc.status == HTTPStatus.BAD_REQUEST
    assert str(exc) == f"{STATUS_EXPLANATIONS[HTTPStatus.BAD_REQUEST]}\nDetail: \n"


def test_constraint_violation_empty_field():
    response_body = CONSTRAINT_VIOLATION_BODY_MISSING_FIELD
    exc = VardefClientError(response_body)
    assert exc.status == HTTPStatus.BAD_REQUEST
    assert exc.detail == (
        "\nDetail: "
        "\nUnknown field: Invalid Dapla team"
        "\nUnknown field: must not be empty"
    )



"""Tests for validators for DatadocMetadata class."""

from __future__ import annotations

import datetime
import re
import warnings
from typing import TYPE_CHECKING

import pytest
from datadoc_model.all_optional import model
from pydantic import ValidationError

from dapla_metadata.datasets.model_validation import ObligatoryDatasetWarning
from dapla_metadata.datasets.model_validation import ObligatoryVariableWarning
from dapla_metadata.datasets.utility.constants import OBLIGATORY_METADATA_WARNING

if TYPE_CHECKING:
    from dapla_metadata.datasets.core import Datadoc


@pytest.mark.parametrize(
    ("model_type", "date_from", "date_until", "raises_exception"),
    [
        ("dataset", datetime.date(2024, 1, 1), datetime.date(1980, 10, 1), True),
        ("dataset", datetime.date(1967, 1, 1), datetime.date(1980, 1, 1), False),
        ("variable", datetime.date(1999, 10, 5), datetime.date(1925, 3, 12), True),
        ("variable", datetime.date(2022, 7, 24), datetime.date(2023, 2, 19), False),
        ("dataset", datetime.date(1967, 1, 1), None, False),
        ("variable", datetime.date(1999, 2, 2), datetime.date(1999, 2, 2), False),
    ],
)
def test_write_metadata_document_validate_date_order(
    model_type,
    date_from,
    date_until,
    raises_exception,
    metadata: Datadoc,
):
    if model_type == "dataset":
        metadata.dataset.contains_data_from = date_from
        metadata.dataset.contains_data_until = date_until
    if model_type == "variable":
        for v in metadata.variables:
            v.contains_data_from = date_from
            v.contains_data_until = date_until
    if raises_exception:
        with pytest.raises(
            ValueError,
            match="contains_data_from must be the same or earlier date than contains_data_until",
        ):
            metadata.write_metadata_document()
    else:
        try:
            metadata.write_metadata_document()
        except ValidationError as exc:
            pytest.fail(str(exc))


def test_write_metadata_document_created_date(
    metadata: Datadoc,
):
    metadata.dataset.metadata_created_date = None
    metadata.write_metadata_document()
    assert metadata.dataset.metadata_created_date is not None


@pytest.mark.parametrize(
    ("variable_date", "date_from", "date_until"),
    [
        (None, datetime.date(1967, 1, 1), datetime.date(1980, 1, 1)),
        (
            datetime.date(2022, 2, 2),
            datetime.date(1999, 3, 3),
            datetime.date(2000, 1, 4),
        ),
    ],
)
def test_variables_inherit_dates(
    variable_date,
    date_from,
    date_until,
    metadata: Datadoc,
):
    metadata.dataset.contains_data_from = date_from
    metadata.dataset.contains_data_until = date_until
    for v in metadata.variables:
        v.contains_data_from = variable_date
        v.contains_data_until = variable_date
    metadata.write_metadata_document()
    for v in metadata.variables:
        if variable_date is None:
            assert v.contains_data_from == metadata.dataset.contains_data_from
            assert v.contains_data_until == metadata.dataset.contains_data_until
        else:
            assert v.contains_data_from == variable_date
            assert v.contains_data_until == variable_date


def test_obligatory_metadata_dataset_warning(metadata: Datadoc):
    with pytest.warns(
        ObligatoryDatasetWarning,
        match=OBLIGATORY_METADATA_WARNING,
    ) as record:
        metadata.write_metadata_document()
    all_obligatory_completed = 100
    num_warnings = 2
    if metadata.percent_complete != all_obligatory_completed:
        assert len(record) == num_warnings
        assert issubclass(record[0].category, ObligatoryDatasetWarning)
        assert OBLIGATORY_METADATA_WARNING in str(
            record[0].message,
        )


def test_obligatory_metadata_variables_warning(metadata: Datadoc):
    with pytest.warns(
        ObligatoryVariableWarning,
        match=OBLIGATORY_METADATA_WARNING,
    ) as record:
        metadata.write_metadata_document()
    all_obligatory_completed = 100
    if metadata.percent_complete != all_obligatory_completed and len(record) > 1:
        assert issubclass(record[1].category, ObligatoryVariableWarning)
        if (
            metadata.variables_lookup["pers_id"]
            and metadata.variables_lookup["pers_id"].name is None
        ):
            assert "[{'pers_id': ['name'," in str(
                record[1].message,
            )


def test_obligatory_metadata_dataset_warning_name(metadata: Datadoc):
    metadata.dataset.name = None
    with pytest.warns(
        ObligatoryDatasetWarning,
        match=OBLIGATORY_METADATA_WARNING,
    ) as record:
        metadata.write_metadata_document()
    assert "name" in str(
        record[0].message,
    )
    # Set value 'name' for first time, a Language object is created
    metadata.dataset.name = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(languageCode="nb", languageText="Navnet"),
        ],
    )
    metadata.dataset.description = None
    with pytest.warns(
        ObligatoryDatasetWarning,
        match=OBLIGATORY_METADATA_WARNING,
    ) as record2:
        metadata.write_metadata_document()
    assert "name" not in str(record2[0].message)

    # Remove value for 'name', value for 'name' is no longer 'None', but 'languageText' is None
    metadata.dataset.name = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(languageCode="nb", languageText=""),
        ],
    )
    with pytest.warns(
        ObligatoryDatasetWarning,
        match=OBLIGATORY_METADATA_WARNING,
    ) as record3:
        metadata.write_metadata_document()
    assert "name" in str(record3[0].message)


def test_obligatory_metadata_dataset_warning_description(metadata: Datadoc):
    """Field name 'description' is a special case because it can match other field names like 'version_description'."""
    error_message: str
    missing_obligatory_dataset = ""
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        metadata.write_metadata_document()
        if issubclass(w[0].category, ObligatoryDatasetWarning):
            error_message = str(w[0].message)
    assert re.search(r"\bdescription\b", error_message)

    # Check that field name is removed from warning when value
    metadata.dataset.description = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(languageCode="nb", languageText="Beskrivelse"),
        ],
    )
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        metadata.write_metadata_document()
        if issubclass(w[0].category, ObligatoryDatasetWarning):
            missing_obligatory_dataset = str(w[0].message)
    assert not re.search(r"\bdescription\b", missing_obligatory_dataset)


def test_obligatory_metadata_dataset_warning_multiple_languages(
    metadata: Datadoc,
):
    missing_obligatory_dataset = ""

    metadata.dataset.description = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(languageCode="nb", languageText="Beskrivelse"),
            model.LanguageStringTypeItem(languageCode="en", languageText="Description"),
        ],
    )
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        metadata.write_metadata_document()
        if issubclass(w[0].category, ObligatoryDatasetWarning):
            missing_obligatory_dataset = str(w[0].message)
    assert not re.search(r"\bdescription\b", missing_obligatory_dataset)

    # Remove value for one language
    metadata.dataset.description = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(languageCode="nb", languageText=""),
            model.LanguageStringTypeItem(languageCode="en", languageText="Description"),
        ],
    )
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        metadata.write_metadata_document()
        if issubclass(w[0].category, ObligatoryDatasetWarning):
            missing_obligatory_dataset = str(w[0].message)
    assert not re.search(r"\bdescription\b", missing_obligatory_dataset)

    # Remove value for all languages
    metadata.dataset.description = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(languageCode="nb", languageText=""),
            model.LanguageStringTypeItem(languageCode="en", languageText=""),
        ],
    )
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        metadata.write_metadata_document()
        if issubclass(w[0].category, ObligatoryDatasetWarning):
            missing_obligatory_dataset = str(w[0].message)
    assert re.search(r"\bdescription\b", missing_obligatory_dataset)


def test_obligatory_metadata_variables_warning_name(metadata: Datadoc):
    variable_with_name = "{'pers_id': ['name'"
    with pytest.warns(
        ObligatoryVariableWarning,
        match=OBLIGATORY_METADATA_WARNING,
    ) as record:
        metadata.write_metadata_document()
    assert metadata.variables_lookup["pers_id"] is not None
    assert metadata.variables_lookup["pers_id"].name is None
    assert variable_with_name in str(record[1].message)

    metadata.variables_lookup["pers_id"].name = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(languageCode="nb", languageText="Navnet"),
        ],
    )
    with pytest.warns(
        ObligatoryVariableWarning,
        match=OBLIGATORY_METADATA_WARNING,
    ) as record2:
        metadata.write_metadata_document()
    assert variable_with_name not in str(record2[1].message)


def test_obligatory_metadata_variables_warning_pseudonymization(metadata: Datadoc):
    metadata.variables_lookup["pers_id"].pseudonymization = model.Pseudonymization(
        pseudonymization_time="2022-10-07T07:35:01Z",
        stable_identifier_type="",
        stable_identifier_version="",
        encryption_algorithm=None,
        encryption_algorithm_parameters=[],
    )
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        metadata.write_metadata_document()
        if issubclass(w[1].category, ObligatoryVariableWarning):
            missing_obligatory_dataset = str(w[1].message)
    assert "encryption_algorithm" in str(missing_obligatory_dataset)
